/*
PUMCL - Parser
used to parse a subset of plantUML (class diagram syntax)
the syntax is slightly different because not all features are supported
and to supported somehow types the syntax is changed e.g. enclose text in "..." to
distinguish it from identifiers and keywords ...

*/
/* lexical grammar */
/*
hints: use left recursive rules (according to http://dinosaur.compilertools.net/bison/bison_6.html#SEC42)
*/

%lex

%x inSingleLineComment inMultiLineComment

%%

<<EOF>>               {return 'EOF';}
[\ \t]+         /* skip whitespace */
[\n]+  return 'NL'



<INITIAL>\/\/ { this.begin('inSingleLineComment');}
<inSingleLineComment>[^\n] {}
<inSingleLineComment>\n { this.begin('INITIAL'); return 'NL';}

<INITIAL>\/\* { this.begin('inMultiLineComment'); }
<inMultiLineComment>\*\/ { this.begin('INITIAL'); }
<inMultiLineComment>\n {}
<inMultiLineComment>[^\n] {}


'::' return 'DOUBLE_COLON'
':' return 'COLON'
',' return 'COMMA'

/* not supported --> skip */
'{field}' return /* skip */
'{method}' return /* skip */

({static}|{classifier}) return 'CLASS_MEMBER_MODIFIER_STATIC'
'{abstract}' return 'CLASS_MEMBER_MODIFIER_ABSTRACT'


/* normal connections */
'<--' return 'ARROWED_LINE'
'<..' return 'ARROWED_LINE_DOTTED'
'-->' return 'ARROWED_LINE_RIGHT'
'..>' return 'ARROWED_LINE_RIGHT_DOTTED'
'<|--' return 'GENERALIZATION'
'--|>' return 'GENERALIZATION_RIGHT'
'<|..' return 'GENERALIZATION_DOTTED'
'..|>' return 'GENERALIZATION_RIGHT_DOTTED'
'*--' return 'COMPOSITION'
'*..' return 'COMPOSITION_DOTTED'
'--*' return 'COMPOSITION_RIGHT'
'..*' return 'COMPOSITION_RIGHT_DOTTED'
'o--' return 'AGGREGATION'
'o..' return 'AGGREGATION_DOTTED'
'--o' return 'AGGREGATION_RIGHT'
'..o' return 'AGGREGATION_RIGHT_DOTTED'

'<-' return 'ARROWED_LINE_SIMPLE'
'<.' return 'ARROWED_LINE_DOTTED_SIMPLE'
'->' return 'ARROWED_LINE_RIGHT_SIMPLE'
'.>' return 'ARROWED_LINE_RIGHT_DOTTED_SIMPLE'
'<|-' return 'GENERALIZATION_SIMPLE'
'-|>' return 'GENERALIZATION_RIGHT_SIMPLE'
'<|.' return 'GENERALIZATION_DOTTED_SIMPLE'
'.|>' return 'GENERALIZATION_RIGHT_DOTTED_SIMPLE'
'*-' return 'COMPOSITION_SIMPLE'
'*.' return 'COMPOSITION_DOTTED_SIMPLE'
'-*' return 'COMPOSITION_RIGHT_SIMPLE'
'.*' return 'COMPOSITION_RIGHT_DOTTED_SIMPLE'
'o-' return 'AGGREGATION_SIMPLE'
'o.' return 'AGGREGATION_DOTTED_SIMPLE'
'-o' return 'AGGREGATION_RIGHT_SIMPLE'
'.o' return 'AGGREGATION_RIGHT_DOTTED_SIMPLE'

/* not specific for uml */
'#--' return 'BOXED_LINE'
'#..' return 'BOXED_LINE_DOTTED'
'--#' return 'BOXED_LINE_RIGHT'
'..#' return 'BOXED_LINE_RIGHT_DOTTED'
'x--' return 'XED_LINE'
'x..' return 'XED_LINE_DOTTED'
'--x' return 'XED_LINE_RIGHT'
'..x' return 'XED_LINE_RIGHT_DOTTED'
'}--' return 'TRIDENT_LINE'
'}..' return 'TRIDENT_LINE_DOTTED'
'--{' return 'TRIDENT_LINE_RIGHT'
'..{' return 'TRIDENT_LINE_RIGHT_DOTTED'
'+--' return 'CIRCLE_PLUS_LINE'
'+..' return 'CIRCLE_PLUS_LINE_DOTTED'
'--+' return 'CIRCLE_PLUS_LINE_RIGHT'
'..+' return 'CIRCLE_PLUS_LINE_RIGHT_DOTTED'
'^--' return 'GENERALIZATION_2_LINE'
'^..' return 'GENERALIZATION_2_LINE_DOTTED'
'--^' return 'GENERALIZATION_2_LINE_RIGHT'
'..^' return 'GENERALIZATION_2_LINE_RIGHT_DOTTED'
/* not specific for uml */
'#-' return 'BOXED_LINE_SIMPLE'
'#.' return 'BOXED_LINE_DOTTED_SIMPLE'
'-#' return 'BOXED_LINE_RIGHT_SIMPLE'
'.#' return 'BOXED_LINE_RIGHT_DOTTED_SIMPLE'
'x-' return 'XED_LINE_SIMPLE'
'x.' return 'XED_LINE_DOTTED_SIMPLE'
'-x' return 'XED_LINE_RIGHT_SIMPLE'
'.x' return 'XED_LINE_RIGHT_DOTTED_SIMPLE'
'}-' return 'TRIDENT_LINE_SIMPLE'
'}.' return 'TRIDENT_LINE_DOTTED_SIMPLE'
'-{' return 'TRIDENT_LINE_RIGHT_SIMPLE'
'.{' return 'TRIDENT_LINE_RIGHT_DOTTED_SIMPLE'
'+-' return 'CIRCLE_PLUS_LINE_SIMPLE'
'+.' return 'CIRCLE_PLUS_LINE_DOTTED_SIMPLE'
'-+' return 'CIRCLE_PLUS_LINE_RIGHT_SIMPLE'
'.+' return 'CIRCLE_PLUS_LINE_RIGHT_DOTTED_SIMPLE'
'^-' return 'GENERALIZATION_2_LINE_SIMPLE'
'^.' return 'GENERALIZATION_2_LINE_DOTTED_SIMPLE'
'-^' return 'GENERALIZATION_2_LINE_RIGHT_SIMPLE'
'.^' return 'GENERALIZATION_2_LINE_RIGHT_DOTTED_SIMPLE'

/* this will break all xx--xx definitions if defined before them because prefix... */
'--' return 'LINE'
'..' return 'LINE_DOTTED'

/* after all that starts with a '-' or '.' */
/* '-' see MINUS */
'.' return 'LINE_DOTTED_SIMPLE'

/* brackets */
'(' return 'BRACKET_OPEN'
')' return 'BRACKET_CLOSE'
(\[\])+ return 'ARRAY_BRACKETS'
'{' return 'CURL_BRACKET_OPEN'
'}' return 'CURL_BRACKET_CLOSE'

'=' return 'EQ'

'-' return 'MINUS' /* CLASS_MEMBER_VISIBILITY_PRIVATE */
'#' return 'HASH' /* CLASS_MEMBER_VISIBILITY_PROTECTED */
'~' return 'TILDE' /* CLASS_MEMBER_VISIBILITY_PACKAGE_PRIVATE */
'+' return 'PLUS' /* CLASS_MEMBER_VISIBILITY_PUBLIC */

'<' return 'ANGEL_BRACKET_OPEN' /* RELATION_DIRECTION_LEFT_OR_GENERIC_OPEN */
'>' return 'ANGEL_BRACKET_CLOSE' /* RELATION_DIRECTION_RIGHT_OR_GENERIC_CLOSE */

[+-]?[\d\.]+ return 'NUMBER'

'abstract' return 'ABSTRACT_KEYWORD'
'class' return 'CLASS_KEYWORD'
'enum' return 'ENUM_KEYWORD'
'interface' return 'INTERFACE_KEYWORD'
'package' return 'PACKAGE_KEYWORD'
'namespace' return 'NAMESPACE_KEYWORD'

/* notes */
'note on link' return 'NOTE_KEYWORD_ON_LINK'
'note left on link' return 'NOTE_KEYWORD_LEFT_ON_LINK'
'note right on link' return 'NOTE_KEYWORD_RIGHT_ON_LINK'
'note top on link' return 'NOTE_KEYWORD_TOP_ON_LINK'
'note bottom on link' return 'NOTE_KEYWORD_BOTTOM_ON_LINK'

'note left of' return 'NOTE_KEYWORD_LEFT_OF'
'note right of' return 'NOTE_KEYWORD_RIGHT_OF'
'note top of' return 'NOTE_KEYWORD_TOP_OF'
'note bottom of' return 'NOTE_KEYWORD_BOTTOM_OF'

'note left' return 'NOTE_KEYWORD_LEFT'
'note right' return 'NOTE_KEYWORD_RIGHT'
'note top' return 'NOTE_KEYWORD_TOP'
'note bottom' return 'NOTE_KEYWORD_BOTTOM'

'note' return 'NOTE_KEYWORD'
'of' return 'OF'
'as' return 'AS'

/* this is also better if we want to support . as a line */
'::' return 'NAMESPACE_SEPARATOR' /* namespace separator */
'namespace' return 'NAMESPACE_KEYWORD'
'package' return 'PACKAGE_KEYWORD'

/* allow ? to support java wildcard as type */
([a-zA-Z][\w\?]*|\?)  return 'IDENT'
\"[^"]+\" return 'STRING'

/lex

%% /* language grammar */


start
  : statements EOF {return $1;}
  ;

emptyStatement
  :
  ;

statements
  : NL
  | statement NL
  | statements NL
  | statements statement NL
  ;

statement
  : relationWithTextDirected
  | maybeAbstractClassDefinitionStatement
  | addToClassStatement
  | interfaceDefinitionStatement
  | enumDefinitionStatement
  | noteDefinitionStatement
  | packageDefinitionStatement
  | associationClassDefinitionStatement
  ;


fullQualifiedName
  : IDENT
  | IDENT DOUBLE_COLON fullQualifiedName
  ;

/* --- relations --- */
/*
  X -- Y
  Y *-- Y
  ...
*/
relation
  : fullQualifiedName someLine fullQualifiedName
  | fullQualifiedName STRING someLine fullQualifiedName
  | fullQualifiedName someLine STRING fullQualifiedName
  | fullQualifiedName STRING someLine STRING fullQualifiedName
  ;

/*
  Driver - Car : "drives"
  Driver - Car : <
  Driver - Car : >
  Driver - Car : < "drives"
  Driver - Car : "drives" >
*/
relationWithTextDirected
  : relation
  | relation COLON STRING
  | relation COLON ANGEL_BRACKET_OPEN
  | relation COLON ANGEL_BRACKET_CLOSE
  | relation COLON ANGEL_BRACKET_OPEN STRING
  | relation COLON STRING ANGEL_BRACKET_CLOSE
  ;

/* --- end relations --- */


/* --- classes --- */

type
  : fullQualifiedName
  | fullQualifiedName ARRAY_BRACKETS
  ;

/* applied generic e.g. Tuple<int, string> where we place types for the generics */



appliedGenericList
  : maybeGenericType
  | appliedGenericList COMMA maybeGenericType
  ;

appliedGenericDefinition
  : IDENT ANGEL_BRACKET_OPEN appliedGenericList ANGEL_BRACKET_CLOSE
  | IDENT ANGEL_BRACKET_OPEN appliedGenericList ANGEL_BRACKET_CLOSE ARRAY_BRACKETS
  ;

maybeGenericType
  : type
  | appliedGenericDefinition
  ;


classProperty
  : IDENT
  | IDENT COLON maybeGenericType
  /*| type IDENT /* HUGE  TODO allow { int count} syntax !! this makes the grammar ambiguous or the automaton */
  ;


methodParamsList
    : classProperty
    | methodParamsList COMMA classProperty
    ;

classMethodParamsList
    : BRACKET_OPEN BRACKET_CLOSE
    | BRACKET_OPEN methodParamsList BRACKET_CLOSE
    ;

/* defining genric e.g. map<T : int,U : IEnumerable<int>>() */

genericDefinition
  : IDENT
  | IDENT COLON maybeGenericType
  ;

genericDefinitionList
  : genericDefinition
  | genericDefinitionList COMMA genericDefinition
  ;

genericsDefinition
  :  ANGEL_BRACKET_OPEN genericDefinitionList ANGEL_BRACKET_CLOSE
  ;

/* end defining genric */


/* HUGE  TODO !! allow: void foo() syntax */
classMethod
  : IDENT classMethodParamsList
  | IDENT classMethodParamsList COLON maybeGenericType

  | IDENT genericsDefinition classMethodParamsList
  | IDENT genericsDefinition classMethodParamsList COLON maybeGenericType
  ;


/*
private
protected
package private
public
*/
classMemberVisibilityModifier
  : MINUS
  | HASH
  | TILDE
  | PLUS
  ;

classMemberCommonModifier
    : CLASS_MEMBER_MODIFIER_STATIC
    | CLASS_MEMBER_MODIFIER_ABSTRACT
    ;

classMemberModifiers
  : classMemberVisibilityModifier
  | classMemberVisibilityModifier classMemberCommonModifier
  | classMemberCommonModifier
  ;

classMember
  : classProperty
  | classMethod
  | classMemberModifiers classProperty
  | classMemberModifiers classMethod
  ;

classMemberList
  : NL
  | classMember NL
  | classMemberList NL
  | classMemberList classMember NL
  ;


stereotypeDefinition
  : ANGEL_BRACKET_OPEN ANGEL_BRACKET_OPEN IDENT ANGEL_BRACKET_CLOSE ANGEL_BRACKET_CLOSE
  ;


nameDefinition
  : fullQualifiedName
  | STRING AS fullQualifiedName
  | fullQualifiedName AS STRING
  ;

nameAndMaybeGenericsDefinitionAndMaybeStereotypeDefinition
  : nameDefinition

/* TODO MAYBE inspect why we can't change this to
  | nameDefinition stereotypeDefinition

  we have to use the explicte 3 rules below...
*/
  | fullQualifiedName stereotypeDefinition
  | STRING AS fullQualifiedName stereotypeDefinition
  | fullQualifiedName AS STRING stereotypeDefinition

  | fullQualifiedName genericsDefinition
  | STRING AS fullQualifiedName genericsDefinition
  | fullQualifiedName genericsDefinition AS STRING

  | fullQualifiedName genericsDefinition stereotypeDefinition
  | STRING AS fullQualifiedName genericsDefinition stereotypeDefinition
  | fullQualifiedName genericsDefinition AS STRING stereotypeDefinition
  ;

noClassBody :;

classBodyDefinition
  : noClassBody
  | CURL_BRACKET_OPEN NL CURL_BRACKET_CLOSE
  | CURL_BRACKET_OPEN NL classMemberList CURL_BRACKET_CLOSE
  ;

classDefinition
  : CLASS_KEYWORD nameAndMaybeGenericsDefinitionAndMaybeStereotypeDefinition classBodyDefinition
  ;

addToClassStatement
  : IDENT COLON classMember
  ;

maybeAbstractClassDefinitionStatement
  : classDefinition
  | ABSTRACT_KEYWORD classDefinition
  | ABSTRACT_KEYWORD nameAndMaybeGenericsDefinitionAndMaybeStereotypeDefinition classBodyDefinition
  ;

/* --- end classes --- */


/* --- interfaces --- */

interfaceMember
  : classMethod
  ;

interfaceMemberList
  : NL
  | interfaceMember NL
  | interfaceMemberList NL
  | interfaceMemberList interfaceMember NL
  ;

noInterfaceBody :;

interfaceBodyDefinition
  : noInterfaceBody
  | CURL_BRACKET_OPEN NL CURL_BRACKET_CLOSE
  | CURL_BRACKET_OPEN NL interfaceMemberList CURL_BRACKET_CLOSE
  ;

interfaceDefinitionStatement
  : INTERFACE_KEYWORD nameAndMaybeGenericsDefinitionAndMaybeStereotypeDefinition interfaceBodyDefinition
  ;

/* --- end interfaces --- */

/* --- enums --- */

enumMember
  : IDENT
  | IDENT EQ NUMBER
  | IDENT EQ STRING
  ;

enumMemberList
  : NL
  | enumMember NL
  | enumMemberList NL
  | enumMemberList enumMember NL
  ;

noEnumBody :;

enumBodyDefinition
  : noEnumBody
  | CURL_BRACKET_OPEN NL CURL_BRACKET_CLOSE
  | CURL_BRACKET_OPEN NL enumMemberList CURL_BRACKET_CLOSE
  ;

enumDefinitionStatement
  : ENUM_KEYWORD nameDefinition enumBodyDefinition
  | ENUM_KEYWORD nameDefinition stereotypeDefinition enumBodyDefinition
  ;

/* --- end enums --- */


/* --- notes --- */

noteNameDefinition
  : AS IDENT
  ;


noteKeywordDirectionOf
  : NOTE_KEYWORD_LEFT_OF
  | NOTE_KEYWORD_RIGHT_OF
  | NOTE_KEYWORD_TOP_OF
  | NOTE_KEYWORD_BOTTOM_OF
  ;

noteKeywordDirection
  : NOTE_KEYWORD_LEFT
  | NOTE_KEYWORD_RIGHT
  | NOTE_KEYWORD_TOP
  | NOTE_KEYWORD_BOTTOM
  ;

noteKeywordOnLink
  : NOTE_KEYWORD_ON_LINK
  | NOTE_KEYWORD_LEFT_ON_LINK
  | NOTE_KEYWORD_RIGHT_ON_LINK
  | NOTE_KEYWORD_TOP_ON_LINK
  | NOTE_KEYWORD_BOTTOM_ON_LINK
  ;

noteDefinitionStatement
  : NOTE_KEYWORD STRING
  | NOTE_KEYWORD STRING noteNameDefinition

  | noteKeywordDirectionOf IDENT COLON STRING

  | noteKeywordDirection COLON STRING

  | noteKeywordOnLink COLON STRING

  ;

/* --- end nodes --- */


/* --- package / module --- */

noPackageBody :;


packageMember
  : maybeAbstractClassDefinitionStatement
  | interfaceDefinitionStatement
  | enumDefinitionStatement
  | packageDefinitionStatement
  | associationClassDefinitionStatement
  ;

packageMemberList
  : NL
  | packageMember NL
  | packageMemberList NL
  | packageMemberList packageMember NL
  ;

packageBodyDefinition
  /*: noPackageBody /* empty bodies are not allowed for packages */
  : CURL_BRACKET_OPEN NL CURL_BRACKET_CLOSE
  | CURL_BRACKET_OPEN NL packageMemberList CURL_BRACKET_CLOSE
  ;

packageDefinitionStatement
  : PACKAGE_KEYWORD fullQualifiedName packageBodyDefinition
  | PACKAGE_KEYWORD fullQualifiedName stereotypeDefinition packageBodyDefinition

  ;

/* --- end package --- */

/* namespace */

/* TODO */

/* end namespace */

/* association class */

associationClassDefinitionStatement
  : BRACKET_OPEN fullQualifiedName COMMA fullQualifiedName BRACKET_CLOSE LINE_DOTTED fullQualifiedName
  ;

/* end association class */

/* all lines ... */
someLine
  :LINE
  | LINE_DOTTED
  | ARROWED_LINE
  | ARROWED_LINE_DOTTED
  | ARROWED_LINE_RIGHT
  | ARROWED_LINE_RIGHT_DOTTED
  | GENERALIZATION
  | GENERALIZATION_RIGHT
  | GENERALIZATION_DOTTED
  | GENERALIZATION_RIGHT_DOTTED
  | COMPOSITION
  | COMPOSITION_DOTTED
  | COMPOSITION_RIGHT
  | COMPOSITION_RIGHT_DOTTED
  | AGGREGATION
  | AGGREGATION_DOTTED
  | AGGREGATION_RIGHT
  | AGGREGATION_RIGHT_DOTTED
  | ARROWED_LINE_SIMPLE
  | ARROWED_LINE_DOTTED_SIMPLE
  | ARROWED_LINE_RIGHT_SIMPLE
  | ARROWED_LINE_RIGHT_DOTTED_SIMPLE
  | GENERALIZATION_SIMPLE
  | GENERALIZATION_RIGHT_SIMPLE
  | GENERALIZATION_DOTTED_SIMPLE
  | GENERALIZATION_RIGHT_DOTTED_SIMPLE
  | COMPOSITION_SIMPLE
  | COMPOSITION_DOTTED_SIMPLE
  | COMPOSITION_RIGHT_SIMPLE
  | COMPOSITION_RIGHT_DOTTED_SIMPLE
  | AGGREGATION_SIMPLE
  | AGGREGATION_DOTTED_SIMPLE
  | AGGREGATION_RIGHT_SIMPLE
  | AGGREGATION_RIGHT_DOTTED_SIMPLE

   /* not specific for uml*/
  | BOXED_LINE
  | BOXED_LINE_DOTTED
  | BOXED_LINE_RIGHT
  | BOXED_LINE_RIGHT_DOTTED
  | XED_LINE
  | XED_LINE_DOTTED
  | XED_LINE_RIGHT
  | XED_LINE_RIGHT_DOTTED
  | TRIDENT_LINE
  | TRIDENT_LINE_DOTTED
  | TRIDENT_LINE_RIGHT
  | TRIDENT_LINE_RIGHT_DOTTED
  | CIRCLE_PLUS_LINE
  | CIRCLE_PLUS_LINE_DOTTED
  | CIRCLE_PLUS_LINE_RIGHT
  | CIRCLE_PLUS_LINE_RIGHT_DOTTED
  | GENERALIZATION_2_LINE
  | GENERALIZATION_2_LINE_DOTTED
  | GENERALIZATION_2_LINE_RIGHT
  | GENERALIZATION_2_LINE_RIGHT_DOTTED
  | BOXED_LINE_SIMPLE
  | BOXED_LINE_DOTTED_SIMPLE
  | BOXED_LINE_RIGHT_SIMPLE
  | BOXED_LINE_RIGHT_DOTTED_SIMPLE
  | XED_LINE_SIMPLE
  | XED_LINE_DOTTED_SIMPLE
  | XED_LINE_RIGHT_SIMPLE
  | XED_LINE_RIGHT_DOTTED_SIMPLE
  | TRIDENT_LINE_SIMPLE
  | TRIDENT_LINE_DOTTED_SIMPLE
  | TRIDENT_LINE_RIGHT_SIMPLE
  | TRIDENT_LINE_RIGHT_DOTTED_SIMPLE
  | CIRCLE_PLUS_LINE_SIMPLE
  | CIRCLE_PLUS_LINE_DOTTED_SIMPLE
  | CIRCLE_PLUS_LINE_RIGHT_SIMPLE
  | CIRCLE_PLUS_LINE_RIGHT_DOTTED_SIMPLE
  | GENERALIZATION_2_LINE_SIMPLE
  | GENERALIZATION_2_LINE_DOTTED_SIMPLE
  | GENERALIZATION_2_LINE_RIGHT_SIMPLE
  | GENERALIZATION_2_LINE_RIGHT_DOTTED_SIMPLE
  ;
