/*
PUMCL - Parser
used to parse a subset of plantUML (class diagram syntax)
the syntax is slightly different because not all features are supported
and to supported somehow types the syntax is changed e.g. enclose text in "..." to
distinguish it from identifiers and keywords ...
*/

/* lexical grammar */
/*
hints:
- we used error recovery (http://dinosaur.compilertools.net/bison/bison_9.html#SEC81)
  to recover from errors in some states (e.g. statements, class members, ...) see all
  rules with special token "error"
- use left recursive rules (according to http://dinosaur.compilertools.net/bison/bison_6.html#SEC42)
*/

%lex

%x inSingleLineComment inMultiLineComment

%%

<<EOF>>               {return 'EOF';}
[\ \t]+         /* skip whitespace */
[\n]+  return 'NL'



<INITIAL>\/\/ { this.begin('inSingleLineComment');}
<inSingleLineComment>[^\n] {}
<inSingleLineComment>\n { this.begin('INITIAL'); return 'NL';}

<INITIAL>\/\* { this.begin('inMultiLineComment'); }
<inMultiLineComment>\*\/ { this.begin('INITIAL'); }
<inMultiLineComment>\n {}
<inMultiLineComment>[^\n] {}


'::' return 'DOUBLE_COLON'
':' return 'COLON'
',' return 'COMMA'

/* not supported --> skip */
'{field}' return /* skip */
'{method}' return /* skip */

({static}|{classifier}) return 'CLASS_MEMBER_MODIFIER_STATIC'
'{abstract}' return 'CLASS_MEMBER_MODIFIER_ABSTRACT'


/* normal connections/relations */
'<--' return 'ARROWED_LINE'
'<..' return 'ARROWED_LINE_DOTTED'
'-->' return 'ARROWED_LINE_RIGHT'
'..>' return 'ARROWED_LINE_RIGHT_DOTTED'
'<|--' return 'GENERALIZATION'
'--|>' return 'GENERALIZATION_RIGHT'
'<|..' return 'GENERALIZATION_DOTTED'
'..|>' return 'GENERALIZATION_RIGHT_DOTTED'
'*--' return 'COMPOSITION'
'*..' return 'COMPOSITION_DOTTED'
'--*' return 'COMPOSITION_RIGHT'
'..*' return 'COMPOSITION_RIGHT_DOTTED'
'o--' return 'AGGREGATION'
'o..' return 'AGGREGATION_DOTTED'
'--o' return 'AGGREGATION_RIGHT'
'..o' return 'AGGREGATION_RIGHT_DOTTED'

'<-' return 'ARROWED_LINE_SIMPLE'
'<.' return 'ARROWED_LINE_DOTTED_SIMPLE'
'->' return 'ARROWED_LINE_RIGHT_SIMPLE'
'.>' return 'ARROWED_LINE_RIGHT_DOTTED_SIMPLE'
'<|-' return 'GENERALIZATION_SIMPLE'
'-|>' return 'GENERALIZATION_RIGHT_SIMPLE'
'<|.' return 'GENERALIZATION_DOTTED_SIMPLE'
'.|>' return 'GENERALIZATION_RIGHT_DOTTED_SIMPLE'
'*-' return 'COMPOSITION_SIMPLE'
'*.' return 'COMPOSITION_DOTTED_SIMPLE'
'-*' return 'COMPOSITION_RIGHT_SIMPLE'
'.*' return 'COMPOSITION_RIGHT_DOTTED_SIMPLE'
'o-' return 'AGGREGATION_SIMPLE'
'o.' return 'AGGREGATION_DOTTED_SIMPLE'
'-o' return 'AGGREGATION_RIGHT_SIMPLE'
'.o' return 'AGGREGATION_RIGHT_DOTTED_SIMPLE'

/* not specific for uml */
'#--' return 'BOXED_LINE'
'#..' return 'BOXED_LINE_DOTTED'
'--#' return 'BOXED_LINE_RIGHT'
'..#' return 'BOXED_LINE_RIGHT_DOTTED'
'x--' return 'XED_LINE'
'x..' return 'XED_LINE_DOTTED'
'--x' return 'XED_LINE_RIGHT'
'..x' return 'XED_LINE_RIGHT_DOTTED'
'}--' return 'TRIDENT_LINE'
'}..' return 'TRIDENT_LINE_DOTTED'
'--{' return 'TRIDENT_LINE_RIGHT'
'..{' return 'TRIDENT_LINE_RIGHT_DOTTED'
'+--' return 'CIRCLE_PLUS_LINE'
'+..' return 'CIRCLE_PLUS_LINE_DOTTED'
'--+' return 'CIRCLE_PLUS_LINE_RIGHT'
'..+' return 'CIRCLE_PLUS_LINE_RIGHT_DOTTED'
'^--' return 'GENERALIZATION_2_LINE'
'^..' return 'GENERALIZATION_2_LINE_DOTTED'
'--^' return 'GENERALIZATION_2_LINE_RIGHT'
'..^' return 'GENERALIZATION_2_LINE_RIGHT_DOTTED'
/* not specific for uml */
'#-' return 'BOXED_LINE_SIMPLE'
'#.' return 'BOXED_LINE_DOTTED_SIMPLE'
'-#' return 'BOXED_LINE_RIGHT_SIMPLE'
'.#' return 'BOXED_LINE_RIGHT_DOTTED_SIMPLE'
'x-' return 'XED_LINE_SIMPLE'
'x.' return 'XED_LINE_DOTTED_SIMPLE'
'-x' return 'XED_LINE_RIGHT_SIMPLE'
'.x' return 'XED_LINE_RIGHT_DOTTED_SIMPLE'
'}-' return 'TRIDENT_LINE_SIMPLE'
'}.' return 'TRIDENT_LINE_DOTTED_SIMPLE'
'-{' return 'TRIDENT_LINE_RIGHT_SIMPLE'
'.{' return 'TRIDENT_LINE_RIGHT_DOTTED_SIMPLE'
'+-' return 'CIRCLE_PLUS_LINE_SIMPLE'
'+.' return 'CIRCLE_PLUS_LINE_DOTTED_SIMPLE'
'-+' return 'CIRCLE_PLUS_LINE_RIGHT_SIMPLE'
'.+' return 'CIRCLE_PLUS_LINE_RIGHT_DOTTED_SIMPLE'
'^-' return 'GENERALIZATION_2_LINE_SIMPLE'
'^.' return 'GENERALIZATION_2_LINE_DOTTED_SIMPLE'
'-^' return 'GENERALIZATION_2_LINE_RIGHT_SIMPLE'
'.^' return 'GENERALIZATION_2_LINE_RIGHT_DOTTED_SIMPLE'

/* this will break all xx--xx definitions if defined before them because prefix... */
'--' return 'LINE'
'..' return 'LINE_DOTTED'

/* after all that starts with a '-' or '.' */
/* '-' see MINUS */
'.' return 'LINE_DOTTED_SIMPLE'

/* brackets */
'(' return 'BRACKET_OPEN'
')' return 'BRACKET_CLOSE'
(\[\])+ return 'ARRAY_BRACKETS'
'{' return 'CURL_BRACKET_OPEN'
'}' return 'CURL_BRACKET_CLOSE'

'=' return 'EQ'

'-' return 'MINUS' /* CLASS_MEMBER_VISIBILITY_PRIVATE */
'#' return 'HASH' /* CLASS_MEMBER_VISIBILITY_PROTECTED */
'~' return 'TILDE' /* CLASS_MEMBER_VISIBILITY_PACKAGE_PRIVATE */
'+' return 'PLUS' /* CLASS_MEMBER_VISIBILITY_PUBLIC */

'<' return 'ANGEL_BRACKET_OPEN' /* RELATION_DIRECTION_LEFT_OR_GENERIC_OPEN */
'>' return 'ANGEL_BRACKET_CLOSE' /* RELATION_DIRECTION_RIGHT_OR_GENERIC_CLOSE */

[+-]?[\d\.]+ return 'NUMBER'

'abstract' return 'ABSTRACT_KEYWORD'
(static|classifier) return 'STATIC_KEYWORD'

'class' return 'CLASS_KEYWORD'
'enum' return 'ENUM_KEYWORD'
'interface' return 'INTERFACE_KEYWORD'
'package' return 'PACKAGE_KEYWORD'
'namespace' return 'NAMESPACE_KEYWORD'

/* notes */
'note on link' return 'NOTE_KEYWORD_ON_LINK'
'note left on link' return 'NOTE_KEYWORD_LEFT_ON_LINK'
'note right on link' return 'NOTE_KEYWORD_RIGHT_ON_LINK'
'note top on link' return 'NOTE_KEYWORD_TOP_ON_LINK'
'note bottom on link' return 'NOTE_KEYWORD_BOTTOM_ON_LINK'

'note left of' return 'NOTE_KEYWORD_LEFT_OF'
'note right of' return 'NOTE_KEYWORD_RIGHT_OF'
'note top of' return 'NOTE_KEYWORD_TOP_OF'
'note bottom of' return 'NOTE_KEYWORD_BOTTOM_OF'

'note left' return 'NOTE_KEYWORD_LEFT'
'note right' return 'NOTE_KEYWORD_RIGHT'
'note top' return 'NOTE_KEYWORD_TOP'
'note bottom' return 'NOTE_KEYWORD_BOTTOM'

'note' return 'NOTE_KEYWORD'
'of' return 'OF'
'as' return 'AS'

/* this is also better if we want to support . as a line */
'::' return 'NAMESPACE_SEPARATOR' /* namespace separator */
'namespace' return 'NAMESPACE_KEYWORD'
'package' return 'PACKAGE_KEYWORD'

/* allow ? to support java wildcard as type */
([a-zA-Z][\w\?]*|\?)  return 'IDENT'
\"[^"]+\" return 'STRING'

/lex

%% /* language grammar */

/*
the model builder functions need to be defined in yy
also yy.model should be the class model
*/

start
  : createStart statements EOF { return $1;}
  ;


createStart
  : { AstHelper.reset(); $$ = AstHelper.model = ModelBuilder.createModel() }
  ;

emptyStatement
  :
  ;

statements
  : NL
  | statement NL { $$ = $1 }
  | statements NL { $$ = $1 }
  | statements statement NL { $$ = $1 }
  | statements error NL { $$ = $1 }
  ;

statement
  : relationWithTextDirected { $$ = ModelBuilder.addRelation(AstHelper.model, $1) }
  | maybeAbstractClassDefinitionStatement { $$ = ModelBuilder.addClass(AstHelper.model, $1) }
  | addToClassStatement
  | interfaceDefinitionStatement { $$ = ModelBuilder.addInterface(AstHelper.model, $1) }
  | enumDefinitionStatement { $$ = ModelBuilder.addEnum(AstHelper.model, $1) }
  | noteDefinitionStatement { $$ = ModelBuilder.addNote(AstHelper.model, $1) }
  | pointingNodeDefinitionStatement { $$ = ModelBuilder.addPointingNote(AstHelper.model, $1) }
  | packageDefinitionStatement { $$ = ModelBuilder.addPackage(AstHelper.model, $1) }
  | associationClassDefinitionStatement { $$ = ModelBuilder.addAssociationClass(AstHelper.model, $1) }
  ;

fullQualifiedName
  : IDENT { $$ = $1 }
  | IDENT DOUBLE_COLON fullQualifiedName { $$ = $1 + Constants.PackageDelimiter + $3 }
  ;

/* --- relations --- */

/*
  X -- Y
  Y *-- Y
  ...
*/
relation
  : fullQualifiedName someLine fullQualifiedName
  { $$ =
    RelationBuilder.createRelation(
           AstHelper.currentRelationId++,
           RelationNodeBuilder.createRelationNode($1, $2.left),
           RelationNodeBuilder.createRelationNode($3, $2.right),
           null,
           $2.isDashed
    )
  }

  | fullQualifiedName STRING someLine fullQualifiedName /* STRING = cardinality */
    { $$ = RelationBuilder.createRelation(
             AstHelper.currentRelationId++,
             RelationNodeBuilder.createRelationNode($1, $3.left, convertString($2)),
             RelationNodeBuilder.createRelationNode($4, $3.right),
             null,
             $3.isDashed
      )
    }

  | fullQualifiedName someLine STRING fullQualifiedName /* STRING = cardinality */
    { $$ = RelationBuilder.createRelation(
             AstHelper.currentRelationId++,
             RelationNodeBuilder.createRelationNode($1, $2.left),
             RelationNodeBuilder.createRelationNode($4, $2.right, convertString($3)),
             null,
             $2.isDashed
      )
    }

  | fullQualifiedName STRING someLine STRING fullQualifiedName /* STRING = cardinality */
    { $$ = RelationBuilder.createRelation(
             AstHelper.currentRelationId++,
             RelationNodeBuilder.createRelationNode($1, $3.left, convertString($2)),
             RelationNodeBuilder.createRelationNode($5, $3.right, convertString($4)),
             null,
             $3.isDashed
      )
    }

  /* named relation ends with X "*" : "name" -- Y */
  | fullQualifiedName STRING COLON STRING someLine STRING fullQualifiedName
    { $$ = RelationBuilder.createRelation(
             AstHelper.currentRelationId++,
             RelationNodeBuilder.createRelationNode($1, $5.left, convertString($2), convertString($4)),
             RelationNodeBuilder.createRelationNode($7, $5.right, convertString($6)),
             null,
             $5.isDashed
      )
    }

  | fullQualifiedName STRING someLine STRING COLON STRING fullQualifiedName
    { $$ = RelationBuilder.createRelation(
             AstHelper.currentRelationId++,
             RelationNodeBuilder.createRelationNode($1, $3.left, convertString($2)),
             RelationNodeBuilder.createRelationNode($7, $3.right, convertString($4), convertString($6)),
             null,
             $3.isDashed
      )
    }

  | fullQualifiedName STRING COLON STRING someLine STRING COLON STRING fullQualifiedName
    { $$ = RelationBuilder.createRelation(
             AstHelper.currentRelationId++,
             RelationNodeBuilder.createRelationNode($1, $5.left, convertString($2), convertString($4)),
             RelationNodeBuilder.createRelationNode($9, $5.right, convertString($6), convertString($8)),
             null,
             $5.isDashed
      )
    }

  /* named relation with visibility modifier X "*" : + "name" -- Y  */
  | fullQualifiedName STRING COLON classMemberVisibilityModifier STRING someLine STRING fullQualifiedName
    { $$ = RelationBuilder.createRelation(
             AstHelper.currentRelationId++,
             RelationNodeBuilder.createRelationNode($1, $6.left, convertString($2), convertString($5), $4),
             RelationNodeBuilder.createRelationNode($8, $6.right, convertString($7)),
             null,
             $6.isDashed
      )
    }

  | fullQualifiedName STRING someLine STRING COLON classMemberVisibilityModifier STRING fullQualifiedName
    { $$ = RelationBuilder.createRelation(
             AstHelper.currentRelationId++,
             RelationNodeBuilder.createRelationNode($1, $3.left, convertString($2)),
             RelationNodeBuilder.createRelationNode($8, $3.right, convertString($4), convertString($7), $6),
             null,
             $3.isDashed
      )
    }

  | fullQualifiedName STRING COLON classMemberVisibilityModifier STRING someLine STRING COLON classMemberVisibilityModifier STRING fullQualifiedName
    { $$ = RelationBuilder.createRelation(
             AstHelper.currentRelationId++,
             RelationNodeBuilder.createRelationNode($1, $6.left, convertString($2), convertString($5), $4),
             RelationNodeBuilder.createRelationNode($11, $6.right, convertString($7), convertString($10, $9)),
             null,
             $6.isDashed
      )
    }
  ;

/*
  Driver - Car : "drives"
  Driver - Car : <
  Driver - Car : >
  Driver - Car : < "drives"
  Driver - Car : "drives" >
*/
relationWithTextDirected
  : relation { $$ = $1 }
  | relation COLON STRING { $$ = RelationBuilder.setName($1, convertString($3)) }
  | relation COLON ANGEL_BRACKET_OPEN { $$ = RelationBuilder.setRelationDirection($1, RelationDirection.endToStart) }
  | relation COLON ANGEL_BRACKET_CLOSE { $$ = RelationBuilder.setRelationDirection($1, RelationDirection.startToEnd) }

  | relation COLON ANGEL_BRACKET_OPEN STRING
    { $$ =
        RelationBuilder.setName(
          RelationBuilder.setRelationDirection($1, RelationDirection.endToStart),
          convertString($4)
        )
    }
  | relation COLON STRING ANGEL_BRACKET_CLOSE
    { $$ =
        RelationBuilder.setName(
          RelationBuilder.setRelationDirection($1, RelationDirection.startToEnd),
          convertString($3)
        )
    }
  ;

/* --- end relations --- */


/* --- classes --- */

type
  : fullQualifiedName { $$ = TypeBuilder.createType($1) }

  /* [] length = 2 --> dimension 1 ... [][] length 4 --> dim = 2 */
  | fullQualifiedName ARRAY_BRACKETS { $$ = TypeBuilder.createType($1, $2.length / 2) }
  ;

/* applied generic e.g. Tuple<int, string> where we place types for the generics */



appliedGenericList
  : maybeGenericType { $$ = [$1]}
  | appliedGenericList COMMA maybeGenericType { $1.push($3); $$ = $1}
  ;

appliedGenericDefinition
  : IDENT ANGEL_BRACKET_OPEN appliedGenericList ANGEL_BRACKET_CLOSE
    { $$ = TypeBuilder.addManyAppliedGenericDefinitions(
        TypeBuilder.createType($1),
        $3
      )
     }
  | IDENT ANGEL_BRACKET_OPEN appliedGenericList ANGEL_BRACKET_CLOSE ARRAY_BRACKETS
    { $$ = TypeBuilder.addManyAppliedGenericDefinitions(
        TypeBuilder.createType($1, $5.length / 2),
        $3
      )
    }
  ;

maybeGenericType
  : type { $$ = $1 }
  | appliedGenericDefinition { $$ = $1 }
  ;


classProperty
  : IDENT { $$ = PropertyBuilder.createProperty($1) }
  | IDENT COLON maybeGenericType { $$ = PropertyBuilder.createProperty($1, $3) }
  /*| type IDENT /* HUGE  TODO allow { int count} syntax !! this makes the grammar ambiguous or the automaton */
  ;


methodParamsList
    : classProperty { $$ = [$1]}
    | methodParamsList COMMA classProperty { $1.push($3); $$ = $1 }
    ;

classMethodParamsList
    : BRACKET_OPEN BRACKET_CLOSE {$$ = []}
    | BRACKET_OPEN methodParamsList BRACKET_CLOSE {$$ = $2}
    ;

/* defining genric e.g. map<T : int,U : IEnumerable<int>>() */

genericDefinition
  : IDENT { $$ = GenericsDefinitionBuilder.createGenericsDefinition($1)}
  | IDENT COLON maybeGenericType { $$ = GenericsDefinitionBuilder.createGenericsDefinition($1, $3)}
  ;

genericDefinitionList
  : genericDefinition { $$ = [$1]}
  | genericDefinitionList COMMA genericDefinition { $1.push($3); $$ = $1}
  ;

genericsDefinition
  :  ANGEL_BRACKET_OPEN genericDefinitionList ANGEL_BRACKET_CLOSE { $$ = $2 }
  ;

/* end defining genric */


/* HUGE  TODO !! allow: void foo() syntax */

/* because InterfaceMethod is the base class for class methods use it here */
classMethod
  : IDENT classMethodParamsList
    { $$ = InterfaceMethodBuilder.addManyParameters(
        InterfaceMethodBuilder.createInterfaceMethod($1, null),
        $2
      )
    }

  | IDENT classMethodParamsList COLON maybeGenericType
    { $$ = InterfaceMethodBuilder.addManyParameters(
        InterfaceMethodBuilder.createInterfaceMethod($1, $4),
        $2
      )
    }

  | IDENT genericsDefinition classMethodParamsList
    { $$ =
      InterfaceMethodBuilder.addManyGenericDefinitions(
         InterfaceMethodBuilder.addManyParameters(
            InterfaceMethodBuilder.createInterfaceMethod($1, null),
            $3
         ),
         $2
      )
    }

  | IDENT genericsDefinition classMethodParamsList COLON maybeGenericType
    { $$ =
      InterfaceMethodBuilder.addManyGenericDefinitions(
         InterfaceMethodBuilder.addManyParameters(
            InterfaceMethodBuilder.createInterfaceMethod($1, $5),
            $3
         ),
         $2
      )
    }
  ;


/*
private
protected
package private
public
*/
classMemberVisibilityModifier
  : MINUS { $$ = VisibilityModifier._private }
  | HASH { $$ = VisibilityModifier._protected }
  | TILDE { $$ = VisibilityModifier._packagePrivate }
  | PLUS { $$ = VisibilityModifier._public }
  ;

classMemberCommonModifier
    : CLASS_MEMBER_MODIFIER_STATIC { $$ = OtherModifier._static }
    | CLASS_MEMBER_MODIFIER_ABSTRACT { $$ = OtherModifier._abstract }
    ;

/* { $$ = { visibilityModifier: null, otherModifier: null} } */
classMemberModifiers
  : classMemberVisibilityModifier { $$ = { visibilityModifier: $1, otherModifier: null} }
  | classMemberVisibilityModifier classMemberCommonModifier { $$ = { visibilityModifier: $1, otherModifier: $2} }
  | classMemberCommonModifier { $$ = { visibilityModifier: null, otherModifier: $1} }
  ;

/* { $$ = {property: , method: } */
classMember
  : classProperty { $$ = {property: $1, method: null} }
  | classMethod
    { $$ = {
        property: null,
        method: MethodBuilder.setVisibilityModifier(
          MethodBuilder.setOtherModifier($1, null), null
        )
      }
    }

  | classMemberModifiers classProperty
    {
      $$ = {
        property: PropertyBuilder.setVisibilityModifier(
          PropertyBuilder.setOtherModifier($2, $1.otherModifier),
          $1.visibilityModifier
        ),
        method: null
      }
    }
  | classMemberModifiers classMethod
    {
      $$ = {
        property: null,
        method: MethodBuilder.setVisibilityModifier(
          MethodBuilder.setOtherModifier($2, $1.otherModifier),
          $1.visibilityModifier
        )
      }
    }
  ;


/* { $$ = {properties: [], methods: []} } */

classMemberList
  : NL { $$ = {properties: [], methods: []} } }

  | classMember NL
    { $$ =
      {
        properties: $1.property !== null ? [$1.property] : [],
        methods: $1.method !== null ? [$1.method] : []
      }
    }
  | classMemberList NL { $$ = $1 }

  | classMemberList classMember NL
    {
      if ($2.property !== null) $1.properties.push($2.property);
      if ($2.method !== null) $1.methods.push($2.method);
      $$ = $1
    }

  | classMemberList error NL { $$ = $1 }
  ;


stereotypeDefinition
  : ANGEL_BRACKET_OPEN ANGEL_BRACKET_OPEN IDENT ANGEL_BRACKET_CLOSE ANGEL_BRACKET_CLOSE { $$ = $3 }
  ;


nameDefinition
  : fullQualifiedName { $$ = {name: $1, alias: null}}
  | STRING AS fullQualifiedName { $$ = {name: $3, alias: convertString($1)}}
  | fullQualifiedName AS STRING { $$ = {name: $1, alias: convertString($3)}}
  ;

nameAndMaybeGenericsDefinitionAndMaybeStereotypeDefinition
  : nameDefinition
    { $$ = {name: $1.name, alias: $1.alias, stereotype: null, genericsDefinitions: []}}

/* TODO MAYBE inspect why we can't change this to
  | nameDefinition stereotypeDefinition

  we have to use the explicte 3 rules below...
*/
  | fullQualifiedName stereotypeDefinition
    { $$ = {name: $1, alias: null, stereotype: $2, genericsDefinitions: []}}

  | STRING AS fullQualifiedName stereotypeDefinition
    { $$ = {name: $3, alias: convertString($1), stereotype: $4, genericsDefinitions: []}}

  | fullQualifiedName AS STRING stereotypeDefinition
    { $$ = {name: $1, alias: convertString($3), stereotype: $4, genericsDefinitions: []}}

  | fullQualifiedName genericsDefinition
    { $$ = {name: $1, alias: null, stereotype: null, genericsDefinitions: $2}}

  | STRING AS fullQualifiedName genericsDefinition
    { $$ = {name: $3, alias: convertString($1), stereotype: null, genericsDefinitions: $4}}

  | fullQualifiedName genericsDefinition AS STRING
    { $$ = {name: $1, alias: convertString($4), stereotype: null, genericsDefinitions: $2}}

  | fullQualifiedName genericsDefinition stereotypeDefinition
    { $$ = {name: $1, alias: null, stereotype: $3, genericsDefinitions: $2}}

  | STRING AS fullQualifiedName genericsDefinition stereotypeDefinition
    { $$ = {name: $3, alias: convertString($1), stereotype: $5, genericsDefinitions: $4}}

  | fullQualifiedName genericsDefinition AS STRING stereotypeDefinition
    { $$ = {name: $1, alias: convertString($4), stereotype: $5, genericsDefinitions: $2}}
  ;

noClassBody :;

classBodyDefinition
  : noClassBody { $$ = {properties: [], methods: []} }
  | CURL_BRACKET_OPEN NL CURL_BRACKET_CLOSE { $$ = {properties: [], methods: []} }
  | CURL_BRACKET_OPEN NL classMemberList CURL_BRACKET_CLOSE { $$ = $3 }
  ;

classDefinition
  : CLASS_KEYWORD nameAndMaybeGenericsDefinitionAndMaybeStereotypeDefinition classBodyDefinition
    { $$ =
        ClassBuilder.addManyMethods(
          ClassBuilder.addManyProperties(
            ClassBuilder.createClass($2.name, $2.stereotype, $2.alias, null, null, $2.genericsDefinitions),
            $3.properties
          ),
          $3.methods
        )
    }
  ;

classMemberCommonModifierNoBraces
  : ABSTRACT_KEYWORD { $$ = OtherModifier._static }
  | STATIC_KEYWORD { $$ = OtherModifier._abstract }
  ;


maybeAbstractClassDefinitionStatement
  : classDefinition
    { $$ = $1 }

  | classMemberCommonModifierNoBraces classDefinition
    { $$ = ClassBuilder.setOtherModifier($2, $1)}

  | classMemberCommonModifierNoBraces nameAndMaybeGenericsDefinitionAndMaybeStereotypeDefinition classBodyDefinition
    { $$ =
        ClassBuilder.addManyMethods(
          ClassBuilder.addManyProperties(
            ClassBuilder.createClass($2.name, $2.stereotype, $2.alias, $1, null, $2.genericsDefinitions),
            $3.properties
          ),
          $.methods
        )
    }
  ;

/* --- end classes --- */

/* --- add to class statement --- */

addToClassStatement
  : IDENT COLON classMember
    { console.warn("TODO addToClassStatement") }
  ;

/* --- end add to class statement --- */

/* --- interfaces --- */

interfaceMember
  : classMethod { $$ = $1 }
  ;

interfaceMemberList
  : NL { $$ = [] }
  | interfaceMember NL { $$ = [$1] }
  | interfaceMemberList NL { $$ = $1 }
  | interfaceMemberList interfaceMember NL { $1.push($2); $$ = $1 }
  | interfaceMemberList error NL { $$ = $1 }
  ;

noInterfaceBody :;

interfaceBodyDefinition
  : noInterfaceBody { $$ = [] }
  | CURL_BRACKET_OPEN NL CURL_BRACKET_CLOSE { $$ = [] }
  | CURL_BRACKET_OPEN NL interfaceMemberList CURL_BRACKET_CLOSE { $$ = $3 }
  ;

/*
nameAndMaybeGenericsDefinitionAndMaybeStereotypeDefinition returns
{ $$ = {name: string, alias: string, stereotype: string, genericsDefinitions: []}}
*/
interfaceDefinitionStatement
  : INTERFACE_KEYWORD nameAndMaybeGenericsDefinitionAndMaybeStereotypeDefinition interfaceBodyDefinition
    { $$ = InterfaceBuilder.addManyMethods(
        InterfaceBuilder.createInterface($2.name, $2.stereotype, $2.alias, null),
        $3
      )
     }
  ;

/* --- end interfaces --- */

/* --- enums --- */

enumMember
  : IDENT
    { $$ = EnumValueBuilder.createEnumValue($1)}
  | IDENT EQ NUMBER
    { $$ = EnumValueBuilder.createEnumValue($1, convertNumber($3))}
  | IDENT EQ STRING
    { $$ = EnumValueBuilder.createEnumValue($1, convertString($3))}
  ;

enumMemberList
  : NL { $$ = [] }
  | enumMember NL { $$ = [$1]}
  | enumMemberList NL { $$ = $1}
  | enumMemberList enumMember NL { $1.push($2); $$ = $1 }
  | enumMemberList error NL { $$ = $1 }
  ;

noEnumBody :;

enumBodyDefinition
  : noEnumBody
    { $$ = []}
  | CURL_BRACKET_OPEN NL CURL_BRACKET_CLOSE
    {$$ = []}
  | CURL_BRACKET_OPEN NL enumMemberList CURL_BRACKET_CLOSE
    { $$ = $3}
  ;

enumDefinitionStatement
  : ENUM_KEYWORD nameDefinition enumBodyDefinition
    { $$ = EnumBuilder.addManyValues(EnumBuilder.createEnum($2.name, null, $2.alias), $3) }

  | ENUM_KEYWORD nameDefinition stereotypeDefinition enumBodyDefinition
    { $$ = EnumBuilder.addManyValues(EnumBuilder.createEnum($2.name, $3, $2.alias), $4) }
  ;

/* --- end enums --- */


/* --- notes --- */

noteNameDefinition
  : AS IDENT
    { $$ = $2}
  ;

noteKeywordDirectionOf
  : NOTE_KEYWORD_LEFT_OF
    { $$ = PointingDirection.left}
  | NOTE_KEYWORD_TOP_OF
    { $$ = PointingDirection.top }
  | NOTE_KEYWORD_RIGHT_OF
    { $$ = PointingDirection.right }
  | NOTE_KEYWORD_BOTTOM_OF
    { $$ = PointingDirection.bot }
  ;

noteKeywordDirection
  : NOTE_KEYWORD_LEFT
    { $$ = PointingDirection.left}
  | NOTE_KEYWORD_TOP
    { $$ = PointingDirection.top }
  | NOTE_KEYWORD_RIGHT
    { $$ = PointingDirection.right }
  | NOTE_KEYWORD_BOTTOM
    { $$ = PointingDirection.bot }
  ;

noteKeywordOnLink
  : NOTE_KEYWORD_ON_LINK
  | NOTE_KEYWORD_LEFT_ON_LINK
  | NOTE_KEYWORD_RIGHT_ON_LINK
  | NOTE_KEYWORD_TOP_ON_LINK
  | NOTE_KEYWORD_BOTTOM_ON_LINK
  ;


noteDefinitionStatement
  : NOTE_KEYWORD STRING
    { $$ = NoteBuilder.createNote(convertString($2)) }
  | NOTE_KEYWORD STRING noteNameDefinition
    { $$ = NoteBuilder.createNote(convertString($2), $3) }
  | noteKeywordOnLink COLON STRING
    { $$ = NoteBuilder.createNote(convertString($3), null, (AstHelper.currentRelationId-1)) }
  ;

pointingNodeDefinitionStatement
  : noteKeywordDirectionOf IDENT COLON STRING
    { $$ = PointingNoteBuilder.createPointingNote(convertString($4), $2, $1)}
  | noteKeywordDirection COLON STRING
    { $$ = PointingNoteBuilder.createPointingNote(convertString($3), AstHelper.lastKnownEntityName, $1)}
  ;


/* --- end nodes --- */

/* TODO */

/* --- package / module --- */

noPackageBody :;

/* { $$ = { _class: {}/null, _interface: {}/null, _enum: {}/null, _associationClass: {}/null, _subPackage: {}/null} } */
packageMember
  : maybeAbstractClassDefinitionStatement
    { $$ = { _class: $1, _interface: null, _enum: null, _associationClass: null, _subPackage: null} }
  | interfaceDefinitionStatement
    { $$ = { _class: null, _interface: $1, _enum: null, _associationClass: null, _subPackage: null} }
  | enumDefinitionStatement
    { $$ = { _class: null, _interface: null, _enum: $1, _associationClass: null, _subPackage: null} }
  | associationClassDefinitionStatement
    { $$ = { _class: null, _interface: null, _enum: null, _associationClass: $1, _subPackage: null} }
  | packageDefinitionStatement
      { console.log("x");
      $$ = { _class: null, _interface: null, _enum: null, _associationClass: null, _subPackage: $1} }
  ;

/* { $$ = { classes: [], interfaces: [], enums: [], associationClasses: [], subPackages: []} } */
packageMemberList
  : NL
    { $$ = { classes: [], interfaces: [], enums: [], associationClasses: [], subPackages: []} }

  | packageMember NL
    { $$ = {
        classes: $1._class !== null ? [$1._class] : [],
        interfaces: $1._interface !== null ? [$1._interface] : [],
        enums: $1._enum !== null ? [$1._enum] : [],
        associationClasses: $1._associationClass !== null ? [$1._associationClass] : [],
        subPackages: $1._subPackage !== null ? [$1._subPackage] : []
      }
    }

  | packageMemberList NL { $$ = $1 }
  | packageMemberList packageMember NL
    {
      if ($2._class !== null) $1.classes.push($2._class);
      if ($2._interface !== null) $1.interfaces.push($2._interface);
      if ($2._enum !== null) $1.enums.push($2._enum);
      if ($2._associationClass !== null) $1.associationClasses.push($2._associationClass);
      if ($2._subPackage !== null) $1.subPackages.push($2._subPackage);
      $$ = $1
    }
  ;

/* { $$ = { classes: [], interfaces: [], enums: [], associationClasses: [], subPackages: []} } */
packageBodyDefinition
  : noPackageBody { $$ = { classes: [], interfaces: [], enums: [], associationClasses: [], subPackages: []} }
  | CURL_BRACKET_OPEN NL CURL_BRACKET_CLOSE { $$ = { classes: [], interfaces: [], enums: [], associationClasses: [], subPackages: []} }
  | CURL_BRACKET_OPEN NL packageMemberList CURL_BRACKET_CLOSE { $$ = $3 }
  ;

packageDefinitionStatement
  : PACKAGE_KEYWORD fullQualifiedName packageBodyDefinition
    { $$ = PackageBuilder.createPackage($2, null, null, $3.classes, $3.interfaces, $3.enums, $3.associationClasses, $3.subPackages) }

  | PACKAGE_KEYWORD fullQualifiedName stereotypeDefinition packageBodyDefinition
    { $$ = PackageBuilder.createPackage($2, null, $3, $4.classes, $4.interfaces, $4.enums, $4.associationClasses, $4.subPackages) }

  ;

/* --- end package --- */

/* namespace */

/* TODO */

/* end namespace */

/* association class */

associationClassDefinitionStatement
  : BRACKET_OPEN fullQualifiedName COMMA fullQualifiedName BRACKET_CLOSE LINE_DOTTED fullQualifiedName
    { $$ =
      AssociationClassBuilder.createAssociationClass($7, (AstHelper.currentRelationId-1), $2, $4)
    }
  ;

/* end association class */

/* all line types ... */

/*  { $$ = {left: LinkKind.normal, right: LinkKind.normal, isDashed: bool} } */
someLine
  : LINE { $$ = {left: LinkKind.normal, right: LinkKind.normal, isDashed: false} }
  | LINE_DOTTED { $$ = {left: LinkKind.normal, right: LinkKind.normal, isDashed: true} }

  | ARROWED_LINE { $$ = {left: LinkKind.arrow, right: LinkKind.normal, isDashed: false} }
  | ARROWED_LINE_DOTTED { $$ = {left: LinkKind.arrow, right: LinkKind.normal, isDashed: true} }
  | ARROWED_LINE_RIGHT { $$ = {left: LinkKind.normal, right: LinkKind.arrow, isDashed: false} }
  | ARROWED_LINE_RIGHT_DOTTED { $$ = {left: LinkKind.normal, right: LinkKind.arrow, isDashed: true} }

  | GENERALIZATION { $$ = {left: LinkKind.generalization, right: LinkKind.normal, isDashed: false} }
  | GENERALIZATION_RIGHT { $$ = {left: LinkKind.normal, right: LinkKind.generalization, isDashed: false} }
  | GENERALIZATION_DOTTED { $$ = {left: LinkKind.generalization, right: LinkKind.arrow, isDashed: true} }
  | GENERALIZATION_RIGHT_DOTTED { $$ = {left: LinkKind.normal, right: LinkKind.generalization, isDashed: true} }

  | COMPOSITION { $$ = {left: LinkKind.composition, right: LinkKind.normal, isDashed: false} }
  | COMPOSITION_DOTTED { $$ = {left: LinkKind.composition, right: LinkKind.normal, isDashed: true} }
  | COMPOSITION_RIGHT { $$ = {left: LinkKind.normal, right: LinkKind.composition, isDashed: false} }
  | COMPOSITION_RIGHT_DOTTED { $$ = {left: LinkKind.normal, right: LinkKind.composition, isDashed: true} }

  | AGGREGATION { $$ = {left: LinkKind.aggregation, right: LinkKind.normal, isDashed: false} }
  | AGGREGATION_DOTTED { $$ = {left: LinkKind.aggregation, right: LinkKind.normal, isDashed: true} }
  | AGGREGATION_RIGHT { $$ = {left: LinkKind.normal, right: LinkKind.aggregation, isDashed: false} }
  | AGGREGATION_RIGHT_DOTTED { $$ = {left: LinkKind.normal, right: LinkKind.aggregation, isDashed: true} }

  | ARROWED_LINE_SIMPLE { $$ = {left: LinkKind.arrow, right: LinkKind.normal, isDashed: false} }
  | ARROWED_LINE_DOTTED_SIMPLE { $$ = {left: LinkKind.arrow, right: LinkKind.normal, isDashed: true} }
  | ARROWED_LINE_RIGHT_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.arrow, isDashed: false} }
  | ARROWED_LINE_RIGHT_DOTTED_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.arrow, isDashed: true} }

  | GENERALIZATION_SIMPLE { $$ = {left: LinkKind.generalization, right: LinkKind.normal, isDashed: false} }
  | GENERALIZATION_RIGHT_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.generalization, isDashed: false} }
  | GENERALIZATION_DOTTED_SIMPLE { $$ = {left: LinkKind.generalization, right: LinkKind.arrow, isDashed: true} }
  | GENERALIZATION_RIGHT_DOTTED_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.generalization, isDashed: true} }

  | COMPOSITION_SIMPLE { $$ = {left: LinkKind.composition, right: LinkKind.normal, isDashed: false} }
  | COMPOSITION_DOTTED_SIMPLE { $$ = {left: LinkKind.composition, right: LinkKind.normal, isDashed: true} }
  | COMPOSITION_RIGHT_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.composition, isDashed: false} }
  | COMPOSITION_RIGHT_DOTTED_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.composition, isDashed: true} }

  | AGGREGATION_SIMPLE { $$ = {left: LinkKind.arrow, right: LinkKind.normal, isDashed: false} }
  | AGGREGATION_DOTTED_SIMPLE { $$ = {left: LinkKind.arrow, right: LinkKind.normal, isDashed: true} }
  | AGGREGATION_RIGHT_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.arrow, isDashed: false} }
  | AGGREGATION_RIGHT_DOTTED_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.arrow, isDashed: true} }

  /*
  { $$ = {left: LinkKind.normal, right: LinkKind.normal, isDashed: false} }
  */
   /* not specific for uml*/
  | BOXED_LINE { $$ = {left: LinkKind.box, right: LinkKind.normal, isDashed: false} }
  | BOXED_LINE_DOTTED { $$ = {left: LinkKind.box, right: LinkKind.normal, isDashed: true} }
  | BOXED_LINE_RIGHT  { $$ = {left: LinkKind.normal, right: LinkKind.box, isDashed: false} }
  | BOXED_LINE_RIGHT_DOTTED { $$ = {left: LinkKind.normal, right: LinkKind.box, isDashed: true} }

  | XED_LINE { $$ = {left: LinkKind.cross, right: LinkKind.normal, isDashed: false} }
  | XED_LINE_DOTTED { $$ = {left: LinkKind.cross, right: LinkKind.normal, isDashed: true} }
  | XED_LINE_RIGHT { $$ = {left: LinkKind.normal, right: LinkKind.cross, isDashed: false} }
  | XED_LINE_RIGHT_DOTTED { $$ = {left: LinkKind.normal, right: LinkKind.cross, isDashed: true} }

  | TRIDENT_LINE { $$ = {left: LinkKind.trident, right: LinkKind.normal, isDashed: false} }
  | TRIDENT_LINE_DOTTED { $$ = {left: LinkKind.trident, right: LinkKind.normal, isDashed: true} }
  | TRIDENT_LINE_RIGHT { $$ = {left: LinkKind.normal, right: LinkKind.trident, isDashed: false} }
  | TRIDENT_LINE_RIGHT_DOTTED { $$ = {left: LinkKind.normal, right: LinkKind.trident, isDashed: true} }

  | CIRCLE_PLUS_LINE { $$ = {left: LinkKind.circlePlus, right: LinkKind.normal, isDashed: false} }
  | CIRCLE_PLUS_LINE_DOTTED { $$ = {left: LinkKind.circlePlus, right: LinkKind.normal, isDashed: true} }
  | CIRCLE_PLUS_LINE_RIGHT { $$ = {left: LinkKind.normal, right: LinkKind.circlePlus, isDashed: false} }
  | CIRCLE_PLUS_LINE_RIGHT_DOTTED { $$ = {left: LinkKind.normal, right: LinkKind.circlePlus, isDashed: true} }

  | GENERALIZATION_2_LINE { $$ = {left: LinkKind.generalization2, right: LinkKind.normal, isDashed: false} }
  | GENERALIZATION_2_LINE_DOTTED { $$ = {left: LinkKind.generalization2, right: LinkKind.normal, isDashed: true} }
  | GENERALIZATION_2_LINE_RIGHT { $$ = {left: LinkKind.normal, right: LinkKind.generalization2, isDashed: false} }
  | GENERALIZATION_2_LINE_RIGHT_DOTTED { $$ = {left: LinkKind.normal, right: LinkKind.generalization2, isDashed: true} }

  | BOXED_LINE_SIMPLE { $$ = {left: LinkKind.box, right: LinkKind.normal, isDashed: false} }
  | BOXED_LINE_DOTTED_SIMPLE { $$ = {left: LinkKind.box, right: LinkKind.normal, isDashed: true} }
  | BOXED_LINE_RIGHT_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.box, isDashed: false} }
  | BOXED_LINE_RIGHT_DOTTED_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.box, isDashed: true} }

  | XED_LINE_SIMPLE { $$ = {left: LinkKind.cross, right: LinkKind.normal, isDashed: false} }
  | XED_LINE_DOTTED_SIMPLE { $$ = {left: LinkKind.cross, right: LinkKind.normal, isDashed: true} }
  | XED_LINE_RIGHT_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.cross, isDashed: false} }
  | XED_LINE_RIGHT_DOTTED_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.cross, isDashed: true} }

  | TRIDENT_LINE_SIMPLE { $$ = {left: LinkKind.trident, right: LinkKind.normal, isDashed: false} }
  | TRIDENT_LINE_DOTTED_SIMPLE { $$ = {left: LinkKind.trident, right: LinkKind.normal, isDashed: true} }
  | TRIDENT_LINE_RIGHT_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.trident, isDashed: false} }
  | TRIDENT_LINE_RIGHT_DOTTED_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.trident, isDashed: true} }

  | CIRCLE_PLUS_LINE_SIMPLE { $$ = {left: LinkKind.circlePlus, right: LinkKind.normal, isDashed: false} }
  | CIRCLE_PLUS_LINE_DOTTED_SIMPLE { $$ = {left: LinkKind.circlePlus, right: LinkKind.normal, isDashed: true} }
  | CIRCLE_PLUS_LINE_RIGHT_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.circlePlus, isDashed: false} }
  | CIRCLE_PLUS_LINE_RIGHT_DOTTED_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.circlePlus, isDashed: true} }

  | GENERALIZATION_2_LINE_SIMPLE { $$ = {left: LinkKind.generalization2, right: LinkKind.normal, isDashed: false} }
  | GENERALIZATION_2_LINE_DOTTED_SIMPLE { $$ = {left: LinkKind.generalization2, right: LinkKind.normal, isDashed: true} }
  | GENERALIZATION_2_LINE_RIGHT_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.generalization2, isDashed: false} }
  | GENERALIZATION_2_LINE_RIGHT_DOTTED_SIMPLE { $$ = {left: LinkKind.normal, right: LinkKind.generalization2, isDashed: true} }
  ;

