/* lexical grammar */
/*
https://zaa.ch/jison/demos/calc/
https://stackoverflow.com/questions/31342563/jison-produce-ast-node-with-multiple-children-for-and-and-or
https://zaa.ch/jison/docs/#usage-from-the-command-line

https://stackoverflow.com/questions/31506696/how-to-get-tokens-in-jison
*/

/*
TODO

TEXT = IDENT because we cannot
distinguish text and identifiers...

string ("...") cannot contain " itself
in docs: known keywords cannot be text e.g. note, class, of, as, note left ...

TODO...
#field
#color

resolved:
# field
#color
#001122

*/

%lex



%%
<<EOF>>               {return 'EOF';}
[\ \t]+         /* skip whitespace */
[\n]  return 'NEWLINE'
':' return 'COLON'
',' return 'METHOD_PARAMETER_SEPARATOR'

'{field}' return 'CLASS_MEMBER_MODIFIER_PRE_FIELD'
'{method}' return 'CLASS_MEMBER_MODIFIER_PRE_METHOD'

({static}|{classifier}) return 'CLASS_MEMBER_MODIFIER_STATIC'
'{abstract}' return 'CLASS_MEMBER_MODIFIER_ABSTRACT'

/*
\#([0-9]{6}|[a-z]+) return 'COLOR'
*/

'<--' return 'OP_ARROWED_LINE'
'<..' return 'OP_ARROWED_LINE_DOTTED'
'-->' return 'OP_ARROWED_LINE_RIGHT'
'..>' return 'OP_ARROWED_LINE_RIGHT_DOTTED'
'<|--' return 'OP_GENERALIZATION'
'--|>' return 'OP_GENERALIZATION_RIGHT'
'<|..' return 'OP_GENERALIZATION_DOTTED'
'..|>' return 'OP_GENERALIZATION_RIGHT_DOTTED'
'*--' return 'OP_COMPOSITION'
'*..' return 'OP_COMPOSITION_DOTTED'
'--*' return 'OP_COMPOSITION_RIGHT'
'..*' return 'OP_COMPOSITION_RIGHT_DOTTED'
'o--' return 'OP_AGGREGATION'
'o..' return 'OP_AGGREGATION_DOTTED'
'--o' return 'OP_AGGREGATION_RIGHT'
'..o' return 'OP_AGGREGATION_RIGHT_DOTTED'
'(' return 'BRACKET_OPEN'
')' return 'BRACKET_CLOSE'
(\[\])+ return 'ARRAY_BRACKETS'
'=' return 'OP_EQ'



'class' return 'CLASS_KEYWORD'
'abstract' return 'ABSTRACT_KEYWORD'
'enum' return 'ENUM_KEYWORD'

'interface' return 'INTERFACE_KEYWORD'

'{' return 'CURL_BRACKET_OPEN'
'}' return 'CURL_BRACKET_CLOSE'

'note on link' return 'NOTE_KEYWORD_ON_LINK'
'note left on link' return 'NOTE_KEYWORD_LEFT_ON_LINK'
'note right on link' return 'NOTE_KEYWORD_RIGHT_ON_LINK'
'note top on link' return 'NOTE_KEYWORD_TOP_ON_LINK'
'note bottom on link' return 'NOTE_KEYWORD_BOTTOM_ON_LINK'

'note left of' return 'NOTE_KEYWORD_LEFT_OF'
'note right of' return 'NOTE_KEYWORD_RIGHT_OF'
'note top of' return 'NOTE_KEYWORD_TOP_OF'
'note bottom of' return 'NOTE_KEYWORD_BOTTOM_OF'

'note left' return 'NOTE_KEYWORD_LEFT'
'note right' return 'NOTE_KEYWORD_RIGHT'
'note top' return 'NOTE_KEYWORD_TOP'
'note bottom' return 'NOTE_KEYWORD_BOTTOM'

'note' return 'NOTE_KEYWORD'
'of' return 'OF'
'as' return 'AS'

/* need to specify after e.g. --> or ..> ...  */
'--' return 'OP_LINE'
'..' return 'OP_LINE_DOTTED'

/* after -- because -*/
'-' return 'CLASS_MEMBER_VISIBILITY_PRIVATE'
'#' return 'CLASS_MEMBER_VISIBILITY_PROTECTED'
'~' return 'CLASS_MEMBER_VISIBILITY_PACKAGE_PRIVATE'
'+' return 'CLASS_MEMBER_VISIBILITY_PUBLIC'

/* specify after tokens with same prefix... e.g. <-- */
'<' return 'RELATION_DIRECTION_LEFT_OR_GENERIC_OPEN'
'>' return 'RELATION_DIRECTION_RIGHT_OR_GENERIC_CLOSE'


/* needs to be after o-- else we got IDENT for o */
/* add ? to support java generics... */
[\w\?]+  return 'TEXT'
\"[^"]+\" return 'STRING'
\"([0-9]+|[0-9]+(\.\.)[0-9]+|[a-zA-Z0-9]+|\*)\" return 'CARDINALITY'




/* e.g. <--x not implemented */


/lex

%% /* language grammar */
start
    : statements EOF
        {return $1;}
    ;

noStatement
    :
    ;

statements
    : statement {$$ = [$1];}
    | statement NEWLINE statements {$$ = [$1, $3];}
    | noStatement
    | NEWLINE statements

    | someAbstractClassDefinition {$$ = [$1];}
    | someAbstractClassDefinition NEWLINE statements {$$ = [$1, $3];}

    | someNote
    | someNote NEWLINE statements {$$ = [$1, $3];}

    | someRelationNote
    | someRelationNote NEWLINE statements { $$ = [$1, $3];}

    | interfaceDefinition {$$ = [$1];}
    | interfaceDefinition NEWLINE statements {$$ = [$1, $3];}

    | enumDefinition {$$ = [$1];}
    | enumDefinition NEWLINE statements {$$ = [$1, $3];}
    ;

someNote
    : NOTE_KEYWORD STRING
    | NOTE_KEYWORD STRING AS TEXT /* TEXT = IDENT */
    | NOTE_KEYWORD_LEFT COLON STRING /* connected to the last class */
    | NOTE_KEYWORD_RIGHT COLON STRING /* connected to the last class */
    | NOTE_KEYWORD_TOP COLON STRING /* connected to the last class */
    | NOTE_KEYWORD_BOTTOM COLON STRING /* connected to the last class */

    /* TEXT = IDENT */
    | NOTE_KEYWORD_LEFT_OF TEXT COLON STRING
    | NOTE_KEYWORD_RIGHT_OF TEXT COLON STRING
    | NOTE_KEYWORD_TOP_OF TEXT COLON STRING
    | NOTE_KEYWORD_BOTTOM_OF TEXT COLON STRING
    ;

/*
noColor
    :
    ;

someColor
    : noColor
    | COLOR
    ;

someRelationNote
    : NOTE_KEYWORD_ON_LINK someColor COLON STRING
    | NOTE_KEYWORD_LEFT_ON_LINK someColor COLON STRING
    | NOTE_KEYWORD_RIGHT_ON_LINK someColor COLON STRING
    | NOTE_KEYWORD_TOP_ON_LINK someColor COLON STRING
    | NOTE_KEYWORD_BOTTOM_ON_LINK someColor COLON STRING
    ;
*/

someRelationNote
    : NOTE_KEYWORD_ON_LINK COLON STRING
    | NOTE_KEYWORD_LEFT_ON_LINK COLON STRING
    | NOTE_KEYWORD_RIGHT_ON_LINK COLON STRING
    | NOTE_KEYWORD_TOP_ON_LINK COLON STRING
    | NOTE_KEYWORD_BOTTOM_ON_LINK COLON STRING
    ;

someOp
    : OP_LINE
    | OP_LINE_DOTTED
    | OP_ARROWED_LINE
    | OP_ARROWED_LINE_DOTTED
    | OP_ARROWED_LINE_RIGHT
    | OP_ARROWED_LINE_RIGHT_DOTTED
    | OP_GENERALIZATION
    | OP_GENERALIZATION_RIGHT
    | OP_GENERALIZATION_DOTTED
    | OP_GENERALIZATION_RIGHT_DOTTED
    | OP_COMPOSITION
    | OP_COMPOSITION_DOTTED
    | OP_COMPOSITION_RIGHT
    | OP_COMPOSITION_RIGHT_DOTTED
    | OP_AGGREGATION
    | OP_AGGREGATION_DOTTED
    | OP_AGGREGATION_RIGHT
    | OP_AGGREGATION_RIGHT_DOTTED
    ;

someOpWithCardinality
    : someOp
    | STRING someOp STRING {$$ = $1 + ' to ' +  $3}
    | someOp STRING
    | STRING someOp
    ;

relationDirection
    : RELATION_DIRECTION_LEFT_OR_GENERIC_OPEN
    | RELATION_DIRECTION_RIGHT_OR_GENERIC_CLOSE
    ;

statement
    /* 1. TEXT = IDENT*/
    : TEXT
    | TEXT someOpWithCardinality TEXT /* if some IDENT is a note only .. as line is allowed!! */
        { $$ = [$1, $2, $3];}
    | TEXT someOpWithCardinality TEXT COLON STRING
        { $$ = [$1, $2, $3, $5];}
    | TEXT someOpWithCardinality TEXT COLON relationDirection
        { $$ = [$1, $2, $3, $5];}
    | TEXT someOpWithCardinality TEXT COLON RELATION_DIRECTION_LEFT_OR_GENERIC_OPEN STRING
        { $$ = [$1, $2, $3, $5, $6];}
    | TEXT someOpWithCardinality TEXT COLON STRING RELATION_DIRECTION_RIGHT_OR_GENERIC_CLOSE
        { $$ = [$1, $2, $3, $5, $6];}
    | TEXT COLON classMember
    ;

type
    : TEXT
    | TEXT ARRAY_BRACKETS
    | TEXT genericTypePart
    | TEXT genericTypePart ARRAY_BRACKETS
    ;

genericTypeList
    : type
    | type METHOD_PARAMETER_SEPARATOR genericTypeList
    ;

/* this is where we place the needed generic type e.g. List<int> so we cannot List<int : Object>*/
/* opposite (defining a generic parameter is done with genericsDefinition */
genericTypePart
    : RELATION_DIRECTION_LEFT_OR_GENERIC_OPEN genericTypeList RELATION_DIRECTION_RIGHT_OR_GENERIC_CLOSE
    ;

/*
data
int data
data : int
 */
classProperty
    : TEXT
    | TEXT TEXT /* TEXT type */
    | TEXT ARRAY_BRACKETS TEXT /* type TEXT */
    | TEXT COLON type
    ;

emptyParams
    :
    ;

paramsList
    : emptyParams classProperty
    | paramsList METHOD_PARAMETER_SEPARATOR classProperty
    ;

classMethodParams
    : BRACKET_OPEN BRACKET_CLOSE
    | BRACKET_OPEN paramsList BRACKET_CLOSE
    ;


emptyGenerics
    :
    ;

oneGenericDefinition
    : TEXT /* TEXT = generic param e.g. T1 */
    | TEXT COLON type /* TEXT = generic param inherit X e.g. T1 : BASE or T1 : BASE<int, ...>  */
    ;

genericsList
    : emptyGenerics oneGenericDefinition
    | genericsList METHOD_PARAMETER_SEPARATOR oneGenericDefinition
    ;

/* here we define a new generic parameter ... e.g. class X<T : IEnumerable> is allowed*
/* opposite (specifying a type for the generic is done with genericTypePart */
genericsDefinition
    : RELATION_DIRECTION_LEFT_OR_GENERIC_OPEN genericsList RELATION_DIRECTION_RIGHT_OR_GENERIC_CLOSE
    ;

/* https://stackoverflow.com/questions/16526051/how-to-represent-generic-parameter-in-uml-method */
classMethodWithGenerics
    : TEXT genericsDefinition classMethodParams
    | TEXT TEXT genericsDefinition classMethodParams /* type IDENT METHOD_BRACKETS e.g. int foo<T>() */
    | TEXT ARRAY_BRACKETS TEXT genericsDefinition classMethodParams /* type[] IDENT METHOD_BRACKETS e.g. int[] foo<T>() */
    | TEXT genericsDefinition classMethodParams COLON type /* 1. TEXT = IDENT  foo<T>() : int */

    | TEXT genericsDefinition TEXT classMethodParams /* 1. TEXT = IDENT (type name) e.g. Generic<T> foo() */
    | TEXT genericsDefinition TEXT genericsDefinition classMethodParams /* 1. TEXT = IDENT (type name) e.g. Generic<T> foo<T>() */
    ;

classMethod
    : classMethodWithGenerics
    | TEXT classMethodParams
    | TEXT TEXT classMethodParams /* type TEXT METHOD_BRACKETS */
    | TEXT ARRAY_BRACKETS TEXT classMethodParams /* type TEXT METHOD_BRACKETS */
    | TEXT classMethodParams COLON type
    ;

classMemberVisibilityModifier
    : CLASS_MEMBER_VISIBILITY_PRIVATE
    | CLASS_MEMBER_VISIBILITY_PROTECTED
    | CLASS_MEMBER_VISIBILITY_PACKAGE_PRIVATE
    | CLASS_MEMBER_VISIBILITY_PUBLIC
    ;

classMemberCommonModifier
    : CLASS_MEMBER_MODIFIER_STATIC
    | CLASS_MEMBER_MODIFIER_ABSTRACT
    ;

classMemberModifiers
    : classMemberVisibilityModifier
    | classMemberVisibilityModifier classMemberCommonModifier
    | classMemberCommonModifier
    ;

classMember
    : noClassMember
    | classProperty
    | CLASS_MEMBER_MODIFIER_PRE_FIELD classProperty
    | classMethod
    | CLASS_MEMBER_MODIFIER_PRE_METHOD classMethod
    ;

noClassMember
    :
    ;


classMembers
    : classMember NEWLINE
    | classMemberModifiers classMember NEWLINE
    | classMember NEWLINE classMembers
    | classMemberModifiers classMember NEWLINE classMembers
    ;

noStereotype
    :
    ;

someStereotype
    : noStereotype
    | RELATION_DIRECTION_LEFT_OR_GENERIC_OPEN RELATION_DIRECTION_LEFT_OR_GENERIC_OPEN TEXT RELATION_DIRECTION_RIGHT_OR_GENERIC_CLOSE RELATION_DIRECTION_RIGHT_OR_GENERIC_CLOSE
    ;

/* all texts are IDENTS */
classDefinition
    : CLASS_KEYWORD TEXT  someStereotype
    | CLASS_KEYWORD TEXT someStereotype CURL_BRACKET_OPEN NEWLINE classMembers CURL_BRACKET_CLOSE
    | CLASS_KEYWORD TEXT someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE

    /* with generics*/
    | CLASS_KEYWORD TEXT genericsDefinition someStereotype
    | CLASS_KEYWORD TEXT genericsDefinition someStereotype CURL_BRACKET_OPEN NEWLINE classMembers CURL_BRACKET_CLOSE
    | CLASS_KEYWORD TEXT genericsDefinition someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE

    /*(re)naming classes syntax */
    | CLASS_KEYWORD TEXT AS STRING someStereotype
    | CLASS_KEYWORD TEXT AS STRING someStereotype CURL_BRACKET_OPEN NEWLINE classMembers CURL_BRACKET_CLOSE
    | CLASS_KEYWORD TEXT AS STRING someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE

    /* with generics*/
    | CLASS_KEYWORD TEXT genericsDefinition AS STRING someStereotype
    | CLASS_KEYWORD TEXT genericsDefinition AS STRING someStereotype CURL_BRACKET_OPEN NEWLINE classMembers CURL_BRACKET_CLOSE
    | CLASS_KEYWORD TEXT genericsDefinition AS STRING someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE

    /*(re)naming classes syntax 2 */
    | CLASS_KEYWORD STRING AS TEXT someStereotype
    | CLASS_KEYWORD STRING AS TEXT someStereotype CURL_BRACKET_OPEN NEWLINE classMembers CURL_BRACKET_CLOSE
    | CLASS_KEYWORD STRING AS TEXT someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE

    /* with generics*/
    | CLASS_KEYWORD STRING AS TEXT genericsDefinition someStereotype
    | CLASS_KEYWORD STRING AS TEXT genericsDefinition someStereotype CURL_BRACKET_OPEN NEWLINE classMembers CURL_BRACKET_CLOSE
    | CLASS_KEYWORD STRING AS TEXT genericsDefinition someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE



    ;

/* all texts are IDENTS */
someAbstractClassDefinition
    : classDefinition
    | ABSTRACT_KEYWORD classDefinition

    | ABSTRACT_KEYWORD TEXT someStereotype
    | ABSTRACT_KEYWORD TEXT someStereotype CURL_BRACKET_OPEN NEWLINE classMembers CURL_BRACKET_CLOSE
    | ABSTRACT_KEYWORD TEXT someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE
    ;

interfaceMember
    : noClassMember
    | classMethod
    ;

/* to be conform to class */
interfaceMembers
    : interfaceMember NEWLINE
    | interfaceMember NEWLINE interfaceMembers
    ;

/* all texts are IDENTS */
interfaceDefinition
    : INTERFACE_KEYWORD TEXT someStereotype
    | INTERFACE_KEYWORD TEXT someStereotype CURL_BRACKET_OPEN NEWLINE interfaceMembers CURL_BRACKET_CLOSE
    | INTERFACE_KEYWORD TEXT someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE

    /* with generics*/
    | INTERFACE_KEYWORD TEXT genericsDefinition someStereotype
    | INTERFACE_KEYWORD TEXT genericsDefinition someStereotype CURL_BRACKET_OPEN NEWLINE interfaceMembers CURL_BRACKET_CLOSE
    | INTERFACE_KEYWORD TEXT genericsDefinition someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE

    /*(re)naming classes */
    | INTERFACE_KEYWORD TEXT  AS STRING someStereotype
    | INTERFACE_KEYWORD TEXT AS STRING someStereotype CURL_BRACKET_OPEN NEWLINE interfaceMembers CURL_BRACKET_CLOSE
    | INTERFACE_KEYWORD TEXT AS STRING someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE

    /* with generics*/
    | INTERFACE_KEYWORD TEXT genericsDefinition AS STRING someStereotype
    | INTERFACE_KEYWORD TEXT genericsDefinition AS STRING someStereotype CURL_BRACKET_OPEN NEWLINE interfaceMembers CURL_BRACKET_CLOSE
    | INTERFACE_KEYWORD TEXT genericsDefinition AS STRING someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE

    /*(re)naming classes syntax 2*/
    | INTERFACE_KEYWORD STRING AS TEXT someStereotype
    | INTERFACE_KEYWORD STRING AS TEXT someStereotype CURL_BRACKET_OPEN NEWLINE interfaceMembers CURL_BRACKET_CLOSE
    | INTERFACE_KEYWORD STRING AS TEXT someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE

    /* with generics*/
    | INTERFACE_KEYWORD STRING AS TEXT genericsDefinition someStereotype
    | INTERFACE_KEYWORD STRING AS TEXT genericsDefinition someStereotype CURL_BRACKET_OPEN NEWLINE interfaceMembers CURL_BRACKET_CLOSE
    | INTERFACE_KEYWORD STRING AS TEXT genericsDefinition someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE
    ;

enumMember
    : noClassMember
    | TEXT
    | TEXT OP_EQ TEXT /* 2. TEXT value */
    | TEXT OP_EQ STRING /* 2. TEXT value */
    ;

enumMembers
    : enumMember NEWLINE
    | enumMember NEWLINE enumMembers
    ;

/* all texts are IDENTS */
enumDefinition
    : ENUM_KEYWORD TEXT someStereotype
    | ENUM_KEYWORD TEXT someStereotype CURL_BRACKET_OPEN NEWLINE enumMembers CURL_BRACKET_CLOSE
    | ENUM_KEYWORD TEXT someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE

    /*(re)naming classes */
    | ENUM_KEYWORD TEXT AS STRING someStereotype
    | ENUM_KEYWORD TEXT AS STRING someStereotype CURL_BRACKET_OPEN NEWLINE enumMembers CURL_BRACKET_CLOSE
    | ENUM_KEYWORD TEXT AS STRING someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE

    /*(re)naming classes syntax 2 */
    | ENUM_KEYWORD STRING AS TEXT someStereotype
    | ENUM_KEYWORD STRING AS TEXT someStereotype CURL_BRACKET_OPEN NEWLINE enumMembers CURL_BRACKET_CLOSE
    | ENUM_KEYWORD STRING AS TEXT someStereotype CURL_BRACKET_OPEN NEWLINE noClassMember CURL_BRACKET_CLOSE
    ;
