/**
 * Created by janis dähne.
 */


class ClassBuilder {
  static createClass(displayName_pk: string,
                     stereotype: string | null = null,
                     alias: string | null = null,
                     otherModifier: OtherModifier | null = null,
                     parentPackage: Package | null = null,
                     genericsDefinitions: GenericsDefinition[] = []): Class {

    return {
      displayName_pk,
      alias,
      parentPackage,
      stereotype,
      methods: [],
      properties: [],
      genericsDefinitions,
      otherModifier
    }
  }

  static setDisplayName(target: Class, displayName_pk: string): Class {
    return set(target, 'displayName_pk', displayName_pk)
  }

  static setStereotype(target: Class, stereotype: string): Class {
    return set(target, 'stereotype', stereotype)
  }

  static setAlias(target: Class, alias: string): Class {
    return set(target, 'alias', alias)
  }

  static setPackage(target: Class, _package: Package): Class {
    return set(target, 'parentPackage', _package)
  }

  static setOtherModifier(target: Class, otherModifier: OtherModifier | null): Class {
    return set(target, 'otherModifier', otherModifier)
  }


  static addGenericsDefinition(target: Class, genericsDefinition: GenericsDefinition): Class {
    target.genericsDefinitions.push(genericsDefinition)
    return target
  }

  static addManyGenericsDefinitions(target: Class, genericsDefinitions: GenericsDefinition[]): Class {
    target.genericsDefinitions = target.genericsDefinitions.concat(genericsDefinitions)
    return target
  }


  static addProperty(target: Class, property: Property): Class {
    target.properties.push(property)
    return target
  }

  static addManyProperties(target: Class, properties: Property[]): Class {
    target.properties = target.properties.concat(properties)
    return target
  }

  static addMethod(target: Class, method: Method): Class {
    target.methods.push(method)
    return target
  }

  static addManyMethods(target: Class, methods: Method[]): Class {
    target.methods = target.methods.concat(methods)
    return target
  }

}
