/**
 * Created by janis dähne.
 */

class InterfaceMethodBuilder {
  static createInterfaceMethod(name: string, returnType: Type | null = null): InterfaceMethod {
    return {
      name_pk: name,
      returnType,
      parameters: [],
      genericsDefinitions: []
    }
  }

  static setMethodName(target: InterfaceMethod, name: string): InterfaceMethod {
    return set(target, 'name_pk', name)
  }

  static setReturnType(target: InterfaceMethod, returnType: Type): InterfaceMethod {
    return set(target, 'returnType', name)
  }

  /**
   * adds a parameter, modifiers are set to null
   * by copy
   * @param {InterfaceMethod} target
   * @param {Property} parameter
   * @returns {InterfaceMethod}
   */
  static addParameter(target: InterfaceMethod, parameter: Property): InterfaceMethod {
    target.parameters.push(PropertyBuilder.createProperty(parameter.name_pk, parameter.type, null, null))
    return target
  }

  /**
   * adds multiple parameters, modifiers are set to null
   *
   * by copy
   * @param {InterfaceMethod} target
   * @param {Property[]} parameters
   * @returns {InterfaceMethod}
   */
  static addManyParameters(target: InterfaceMethod, parameters: Property[]): InterfaceMethod {
    for (let i = 0; i < parameters.length; i++) {
      let parameter = parameters[i];
      target.parameters.push(PropertyBuilder.createProperty(parameter.name_pk, parameter.type, null, null))

    }
    return target
  }

  static addGenericDefinition(target: InterfaceMethod, genericsDefinition: GenericsDefinition): InterfaceMethod {
    target.genericsDefinitions.push(genericsDefinition)
    return target
  }

  static addManyGenericDefinitions(target: InterfaceMethod, genericsDefinitions: GenericsDefinition[]): InterfaceMethod {
    target.genericsDefinitions = target.genericsDefinitions.concat(genericsDefinitions)
    return target
  }
}