/**
 * Created by janis dähne.
 */

declare var classDiagrams: any
const parser2 = classDiagrams
const lexer2 = parser2.lexer

const writeTokens2 = true

interface TestResult {
  program: string
  tokens: string[]
  passed: boolean
  result: any | null,
  error: any | null
}

function runTest(program: string): TestResult {

  let result = null
  let tokens: string[] = []

  if (program[program.length - 1] != '\n') {
    program = program + '\n'
  }

  if (writeTokens2) {
    try {
      lexer2.setInput(program)
      let token = ''
      while (!lexer2.done) {
        token = lexer2.lex();
        /* Look up the token name if necessary */
        if (token in parser2.terminals_) {
          token = parser2.terminals_[token];
        }
        tokens.push('(' + token + ', ' + lexer2.yytext + ')')
      }

    } catch (err) {
      console.error(err)
      return {
        program,
        tokens: tokens,
        passed: false,
        result: null,
        error: err
      }
    }
  }


  try {

    result = parser2.parse(program)
  } catch (err) {
    console.error(err)
    return {
      program,
      tokens: tokens,
      passed: false,
      result: null,
      error: err
    }
  }

  return {
    program,
    tokens: tokens,
    passed: true,
    result: result,
    error: null
  }
}

function displayTestResult(result: TestResult) {
  document.write('<pre>')
  let escaped = result.program
    .replace(/</g, "&lt;")
    .replace(/>/g, "&gt;")
  document.writeln(escaped)
  document.write('</pre>')
  document.writeln(result.passed ? 'OK' : 'ERROR:')
  document.write('<br />')

  if (writeTokens2) {
    document.write('<pre>')

    result.tokens.forEach(token => {
      document.write(token)
      document.write("   ")
    })

    document.write('</pre>')
  }


  if (result.error) {
    document.writeln(result.error)
    document.write('<br />')
  }

  document.write('<br />')
}

