# PUMCL

plantUML Compiler/ plantUML class diagrams

This is a compiler to translate plantUML Class diagrams into images.

![example img not found](example_2017_12_10.png)

---


if *entity* is mentioned it could be a class | interface | enum | association class

### used projects

- jison - https://zaa.ch/jison/docs/ | https://github.com/zaach/jison
- theme - http://ethanschoonover.com/solarized | https://github.com/altercation/solarized
- viz.js - http://viz-js.com | https://github.com/mdaines/viz.js/


### similar projects

- **http://abstratt.github.io/textuml/readme.html** | https://github.com/abstratt/textuml

- http://nomnoml.com | https://github.com/skanaar/nomnoml
- https://mermaidjs.github.io | https://github.com/knsv/mermaid

- more: https://modeling-languages.com/uml-tools/#textual


#### possible other renderers than viz.js

- https://github.com/mstefaniuk/graph-viz-d3-js

- https://github.com/cpettitt/dagre-d3 & d3js


---

## Real Start

### Pipeline

- (use jison to create the parser, only )
- the parser parses the input 
- this creates a model of the diagram (see files in `/model/`) with the functions in `/modelBuilder/`
- post process the model
- the model is then translated via files in `transformers/graphviz/` into hraphviz/dot code
- the dot code is then used as input for viz.js
- result is a png/svg ... all what is supported by viz.js

*shot version:*

Parse plantUML -> model -> create graphviz dot code -> vizjs -> svg

**Actually the syntax is slightly different to plantUML in oder to build a proper compiler/transpiler.**

We **MAYBE** want to check the class diagrams in future versions so we try to keep some more strict syntax (e.g. types) than the original plantUML.


### Differences between PUMCL and plantUML syntax

- the syntax `[type] [name]` is not supported in any rule, use `[name] : [type]`  
  (not sure if this is even possible to parse if both should be supported ... or if it's just me)  
- texts needs to be enclosed in `"..."` in pumcl, plantUML uses only text
- links with `.` or `-` are not supported
- namespaces are not supported
- `{field}, {method}` modifier are not supported **BUT are ignored by the parser** so this should not break anything
- all `skinparam` are not supported
- colors are not supported
- lollipop interface are not supported  
  (because the semantics are not clear ... are there are classes/methods/just text ??)
- change link/relation directions are not supported
  (no time)
- for packages the separator `::` is used and cannot be changed
- advanced class bodys with fancy lines are not supported   
  (no time)
- multiple files are not supported
- hiding is not supported

- pumcl supports comments (single line, multi line)
- pumcl supports relation ends with visibility modifier & names (not only cardinality)
  

So whats left?

See below for all features that are supported by the parser

---

## TODOs

- allow int? --> ? in types e.g. for c# 

- add lollipop interface to parser & transformer 

- update syntax diagram

- how to use/build the parser??

- how to use the transformer...

- list keywords that cannot be used as TEXT

- for type checker: if IDENT has no type then assume class...

- bundle a single file version of pumucl as umd model

- TODO add two node types support for parser (already supported by transformer) e.g. `A <--> B`

- **if relation ends not exists ... error or auto creation??**

#### todo model creation

- gather primitive types (everything that is not an entity)
  - if we allow use before definition this can only be done after model creation...
  - also equal primitive types (equal by name see equality checker) should reference the same js object
- entity types can also be used as types ... usage --> create if not exists or error??
  - add to class statement??
  

#### todo post processing


- check that all relation parts exists (disallow packages??)

- check that association classes are only specified between classes
- check that association class is defined!
- check entity names are unique (with respect to package)
- check if auto generated note names (for anonymous notes) overwrite explicit note names
- check if all used types are defined
  - for primitive types this is ok
  - entity types can also be used as types ... usage --> create if not exists or error??
  - generic parameter types must be defined ether in methods or on entity level

- check that for every association class the two classes have a relation...

## Post Processing

the post processing happens right after the parsing has finished and the model has been created

- all entities names inside packages are replaced by the full qualified name
- for all entities the parentPackage property is set

## Hints

Full qualified names have no generic part.

```text
class X<T>
//not allowed: class X<T, T2> 

class Y

X -- Y //not X<T> !

class M<T> {
  foo: List<T>
  bar: List<T1, T2> //currently allowed 
}

```

## The Parser

The parser is newline sensitive, empty lines are allowed (if not in a statement)

Here is a list of features that are supported by the parser (not by the transformer!)

- [x] comments (c-style comments)
  - [x] single line
  
    ```text
    class X //very important
    ```
    
  - [x] multi line
  
    ```text
    class /* very */ Important /* multi
    line
    comment
    */
    ```

- [x] relations 

  Note that the relation ends needs to be full qualified names.
  If the class X is inside package P and one wants to create a relation to another class Y
  the code would be `X::P -- Y`.
  Relations can only be specified outside of packages.  

  ```text
  ..
  --
  <-- -->
  <|-- --|>
  *-- --*
  o-- --o
  ```
- [x] relations 2

  ```text
  #-- --#
  x-- --x
  }-- --{
  +-- --+
  ^-- --^
  ```
- [x] relations with text

  ```text
  Class1 -- class2 : "text"
  ```
- [x] relations with directions  

  Texts need to be enclosed in `"..."`
  
  ```text
  Driver - Car : "drives cars" > 
  Car *- Wheel : "have" 4 >
  Car *- Wheel : > 
  Car -- Person : < "owns"
  Car -- Person : <
  ```
  
- [x] relation ends  

  Note that for relation ends the cardinality is needed!
  
  - [x] with text
  
    ```text
    `X "*" : "xs" -- "*" Y`,
    `Student "*" -- "*" : "courses" Course`,
    `Student "*": "students" -- "*" : "courses" Course`,
  
    `X "*" : "name" -- "*" Y : "some relation"`,
    `X "*" -- "*" : "name" Y : "some relation"`,
    `X "*": "nameForX" -- "*" : "nameForYFromX" Y : "some relation"`,
    ```
  - [x] with visibility modifier
  
    ```text
    `X  "*" : + "name" -- "*" Y`,
    `X "*" -- "*" : + "name" Y`,
    `X "*": + "name" -- "*" : + "name" Y`,
  
    `X "*" : - "name" -- "*" Y : "some relation"`,
    `X "*" -- "*" : ~ "name" Y : "some relation"`,
    `X "*": #"name" -- "*" : -"name" Y : "some relation"`,
    ```


- [ ] classes
  - [x] basic
  
    ```text
    class Class1
    ```
  - [x] advanced
  
    ```text
    class Class1 {
    }
    class Empty {



    }
    ```
  - [x] inline (declare classes)
  
    ```text
    Class1 -- class2
    ```

  - [x] cardinality
  
    ```text
    Class01 "1" *-- "many" Class02
    Class01 "1" *-- "many" Class02 : "text"
    ```
  - [x] properties (with types)
  
    Note that the type needs to be specified with `: type` after the property if any!
    
    ```text
    Object : hashCode
    Object : hashCode : int
    Object : hashCode : int[]
    Object : hashCode : int[][]
    ```
  - [x] methods (with types, params)
  
    Note that the type needs to be specified with `: type` after the method if any!
    
     ```text
     Object : getHashCode()
     Object : getHashCode() :int
     Object : getHashCode():int[]
     Object : getHashCode()     : int[][]
 
    Object : foo(arg1)
    Object : foo(arg1 : int)
    Object : foo(arg1 : int) : int
    Object : foo(arg1 : int, arg2) : int
    Object : foo(arg1 : int, arg2 : int) : int
    Object : foo(arg1 : int[][], arg2 : int) : int[]
     ```
  - [x] properties in advanced classes
  
    Note that the type needs to be specified with `: type` after methods/fields if any!
    
    ```text
    class X {
        data: string
        method() : void
        method2():void
        method() : bool
    }
    ```
  - [ ] {field}, {method} modifier  
  
    Ignored by the parser
    
    **So better don't use this** 
    
    ```text
    class Dummy { 
        {field} A field (despite parentheses)
        {method} Some method 
    }
    ```
  - [x] field/method visibility  
  
    ```text
    class Dummy {
        -fieldPrivate
        #fieldProtected
        ~methodPackagePrivate()
        +methodPublic() 
    }
    class Dummy {
        - int fieldPrivate
        # fieldProtected : int
        ~ {method} void methodPackagePrivate()
        + {method} methodPublic() : void
    }
    ```
  - [ ] skinparam classAttributeIconSize 0  
  
    *Probably* never supported because this is better done in css??
    
    ```text
    skinparam classAttributeIconSize 0 
    class Dummy { 
    -field1
    }
    ```
  - [x] abstract & static (classifier) modifier
  
    ```text
    class Dummy { 
        {static} id 
        {abstract} methods() 
    }
    class X {
        + {static} {field} foo : int
        - {classifier} {field} foo : int
        ~ {abstract} {field} foo : int
                
        + {static} {method} foo2() : int[]
        + {classifier} {method} foo2() : int[]
        + {abstract} {method} foo2() : int[]
    }
    ```
  - [ ] class body with lines  
  
    *Probably* never supported because not really needed...??
    
  - [x] stereotypes
  
    ```text
    class Object << general >>
    class Object << general >> {
    }
    ```
- [x] notes  

  Because notes can contain new lines and the parser is new line sensitive we decided that every note text needs to be enclosed in `"..."`
  
  - [x] basic
  
    ```text
    note "this is a note"
    note "this is a note" as Note1
    ```
  - [x] connect to classes
  
    ```text
    class object
    note " This is a floating note" as N1

    object .. N1
    N1 .. object
    ```
    Note that only `..` are allowed for connections. This is **NOT** checked by the parser (this is because we cannot distinguish identifiers e.g. class names and text). This is also the reason because note uses strings (which are always enclosed in `"..."`).
    
  - [x] positions
  
    ```text
    class Object << general >>

    note top of Object : "In java, every class\nextends this one."
    note bottom of Object : "comment"
    note left of Object : "comment"
    note right of Object : "comment"

    class Foo
    note left: "On last defined class"
    note right: "On last defined class"
    note top: "On last defined class"
    note bottom: "On last defined class"
    ```
    
    Note that the parser does **NOT** checks if there is a class definition above a note.
    
  - [ ] multi line notes  
  
    **Not needed**  
    Use a normal note. Because we use `"..."` newlines are handled correctly
    
  - [x] notes on relations  
  **Colors are not supported**
  
    ```text
    Dummy --> Foo : "A link"
    note on link: "note that is red"

    note on link: "black is default color?"

    note top on link: "black is default color?"
    note bottom on link: "black is default color?"
    note left on link: "black is default color?"
    note right on link: "black is default color?"

    ```
    
    Note that the parser don't checks if there is really a link/relation above the note!
    
- [x] abstract classes

  ```text
  abstract AbstractCollection
  abstract class AbstractList

  abstract class AbstractList {
     foo() : void
     bar()
  }
  
  ```
- [x] interfaces  

  Interfaces can only contain methods.
  No modifiers are allowed for the methods (Because they are implicit public)! 
  
  ```text
  interface IList
  interface IList {
    foo()
    foo(): void
  }
  interface IList <<stereotype>>
  interface IList<T : IEnumerable<int>> <<stereotype>> {
  }
  ```
- [x] enums  

  Enums allow properties (without modifiers) and also (default) values via `= value`.  
  Stereotypes are allowed too (e.g. to specify the type??).
  
  ```text
  enum Unit
  interface Unit {
   EUR
   dollar
  }
  interface Unit2 {
   m = 0
   dm = 1
   cm
  }
  interface Unit3 <string> {
   m = "meters"
   dm = "decimeter"
   cm = "centimeter "
  }
  ```
- [x] (re)naming
  - [x] classes
  
    ```text
    class " This is my class" as class1 
    class class2 as "It works this way too"

    class " This is my class" as class1 <<optional>> { 
    }
    class class2 as "It works this way too" <<optional>> {
    }
    ```
  - [x] interfaces
  
    ```text
    interface " This is my interface" as interface1
    interface interface2 as " It works this way too"

    interface " This is my interface" as interface1 <<optional>> {
     foo() : void
    }
    interface interface2 as " It works this way too" <<stereotype>> {
      foo() : void
    }
    ```
  - [x] enums
  
    ```text
    enum " This is my enum" as enum1
    enum enum2 as " It works this way too"

    enum " This is my enum" as enum1 <<optional>> {
      DAY
      MONTH
      YEAR
    }
    enum enum2 as " It works this way too" <<stereotype>> {
      DAY = 1
      MONTH = 2
      YEAR = 3
    }
    ```
- [ ] hiding classes, interfaces, members, ...  
     *Probably* never supported because the ui or render should do this

- [x] generics
  - [x] for classes
    - [x] class level
    
      ```text
      class List<T>
      class List<T> {
      }
      class List<T : System::Console> {
      }
      class List<T> <<stereotype>>
      class List<T> as "display text for class"
      class List<T> as "display text" <<stereotype>>
      
      class "display text" as List<T>
      class "display text" as List<T> <<stereotype>>
      class "display text" as List<T> <<stereotype>> {
      }
      
      class List<T, T2, T3>
      class Tuple<T1, T2 : Tuple3<int, string[], List<int>, Pair<string,long>[]>
      ```
    - [x] method level
    
      ```text
      class CollectionHelper <<static>> {
          map<TIn, TOut>(array : Tin[]) : TOut[]
          filter<T>(T[] array, predicate : Func<T, bool>) : T[]
        }
      ```
  - [x] for interfaces
    - [x] interface level
    
      ```text
      interface List<T>
      interface List<T> {
      }
      interface List<T> <<stereotype>>
      interface List<T> as "display text for interface"
      interface List<T> as "display text" <<stereotype>>
      
      interface "display text" as List<T>
      interface "display text" as List<T> <<stereotype>>
      interface "display text" as List<T> <<stereotype>> {
      }
      
      interface List<T, T2, T3>
      interface Tuple<T1, T2 : Tuple3<int, string[], List<int>, Pair<string,long>[]>
      ```
    - [x] method level
    
      ```text
      interface ICollection {
        map<TIn, TOut>(array : Tin[]) : TOut[]
        filter<T>(T[] array, predicate : Func<T, bool>) : T[]
      }
      interface Java8Predicate<T> {
        test(t : T) : boolean 
        negate() : Java8Predicate<T>
        and(other) : Java8Predicate<T>
        or(other : Java8Predicate<? : T>) : Java8Predicate<T>
        {static} isEqual(targetRe) : Java8Predicate<T>
      }
      ```
      
- [x] packages  

  Note that the separator is `::`
  
  ```text
  package foo1::foo2:foo3 {
  }
  package foo1 {
    class X
    package foo2 {
      class Nested<T>
    }
  }
  ```

- [ ] namespaces  

  Not sure what's the difference between package and namespace...
  Even though namespaces has some more syntax to parse ...  
  Use package ;)

- [ ] lollipop interface  

  Not supported, not properly defined... classe/methods/text???  
  
- [ ] change line/relation direction
  
  Not supported (no time)
  
  ```text
  foo -left-> dummyLeft 
  foo -right-> dummyRight 
  foo -up-> dummyUp 
  foo -down-> dummyDown
  ```
  
- [x] association classes  

  ```text
  Student " 0..*" - " 1..*" Course 
  (Student, Course) .. Enrollment
  
  class Enrollment { 
    drop() 
    cancel() 
  }

  Student2 " 0..*" - " 1..*" Course2 
  (Student2, Course2) .. Enrollment
  
  ```
  
  
## The Graphviz Transformer

Not all parsable constructs can be transformed by graphviz ...

*For a feature to be supported the parser must support it first*

Not supported by the transformer:

- [ ] entity level generics (e.g. class List\<T\>)  
      Note that this gets rendered as `xLabel` but one cannot specify a relative position in graphviz for this...
      - for svg we could replace the generics ... but for large lists this would be bad (overlap other elements)
      

- [ ] add to class statement  
      Not implemented yet because we create a new class every time we encounter one...  
      we would need a lookup
- [ ] pointing notes  
      not possible with dot because of shape??
- [ ] notes with `note left/top/right/bottom of ...`  
      not possible with dot because of no control over layout??

  
---
  
#### useful for er diagrams

https://stackoverflow.com/questions/6219933/how-does-one-define-double-lines-for-edge-and-node-shapes-in-graphviz-dot

http://tonyballantyne.com/graphs.html

and half dotted lines:
```dot
digraph ClassDiagram {
        fontname = "Bitstream Vera Sans"
        fontsize = 12

    i [shape=point,width=0,height=0]
    A -> i
    i -> B
}
```