/**
 * Created by janis dähne.
 */

declare var classDiagrams: any
declare var pumcl: any
let parser = pumcl === undefined ? classDiagrams : pumcl //
let lexer = parser.lexer



interface JisonErrorHash {
  /**
   * expected keywords
   */
  expected: string[]
  line: number
  loc: {
    first_column: number
    first_line: number
    last_column: number
    last_line: number
  }
  recoverable: boolean
  text: string
  token: string
}

//if enabled the parser will not fail, only report all errors
//but the test will say OK
const enableErrorRecovery = false

parser.yy.parseError = function(err:string, hash:any): void {
  //console.log(err)
  //console.log(hash)

  if (!enableErrorRecovery) {
    throw err
  } else {
    console.error(err)
  }
}

const writeTokens = false
let count = 0

interface TestResult {
  program: string
  tokens: string[]
  passed: boolean
  result: any | null,
  error: any | null
}

function runTest(program: string): TestResult {

  let result = null
  let tokens: string[] = []

  if (program[program.length - 1] != '\n') {
    program = program + '\n'
  }

  if (writeTokens) {
    try {
      lexer.setInput(program)
      let token = ''
      while (!lexer.done) {
        token = lexer.lex();
        /* Look up the token name if necessary */
        if (token in parser.terminals_) {
          token = parser.terminals_[token];
        }
        tokens.push('(' + token + ', ' + lexer.yytext + ')')
      }

    } catch (err) {
      console.error(err)
      return {
        program,
        tokens: tokens,
        passed: false,
        result: null,
        error: err
      }
    }
  }


  try {

    result = parser.parse(program)
  } catch (err) {
    console.error(err)
    return {
      program,
      tokens: tokens,
      passed: false,
      result: null,
      error: err
    }
  }

  return {
    program,
    tokens: tokens,
    passed: true,
    result: result,
    error: null
  }
}

function displayTestResult(result: TestResult) {

  count++;

  document.write('<pre>')
  let escaped = result.program
    .replace(/</g, "&lt;")
    .replace(/>/g, "&gt;")
  document.writeln(escaped)
  document.write('</pre>')
  document.writeln(result.passed ? 'OK' : 'ERROR:')
  document.write('<br />')

  if (writeTokens) {
    document.write('<pre>')

    result.tokens.forEach(token => {
      document.write(token)
      document.write("   ")
    })

    document.write('</pre>')
  }


  if (result.error) {
    document.writeln(result.error)
    document.write('<br />')
  }

  document.write('<br />')
}

